# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import Callable, Iterable, Mapping, MutableMapping, Sequence
from types import MappingProxyType
from typing import Any, TypeVar

from pyomo.common.enums import ObjectiveSense
from pyomo.common.errors import DeveloperError
from pyomo.common.numeric_types import value
from pyomo.contrib.solver.common.results import SolutionStatus, TerminationCondition

from pyomo.contrib.solver.solvers.knitro.api import knitro
from pyomo.contrib.solver.solvers.knitro.callback import build_callback_handler
from pyomo.contrib.solver.solvers.knitro.package import Package
from pyomo.contrib.solver.solvers.knitro.typing import (
    BoundType,
    Callback,
    ItemData,
    ItemType,
    StructureType,
    ValueType,
)
from pyomo.contrib.solver.solvers.knitro.utils import NonlinearExpressionData
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.base.var import VarData
from pyomo.repn.standard_repn import generate_standard_repn


def parse_bounds(
    items: Iterable[ItemType], idx_map: Mapping[int, int]
) -> Mapping[BoundType, MutableMapping[int, float]]:
    bounds_map = {bnd_type: {} for bnd_type in BoundType}
    for item in items:
        i = idx_map[id(item)]
        if isinstance(item, VarData):
            if item.fixed:
                bounds_map[BoundType.EQ][i] = value(item.value)
                continue
            lb, ub = item.bounds
            if lb is not None:
                bounds_map[BoundType.LO][i] = lb
            if ub is not None:
                bounds_map[BoundType.UP][i] = ub
        elif isinstance(item, ConstraintData):
            lb, _, ub = item.to_bounded_expression(evaluate_bounds=True)
            if item.equality:
                bounds_map[BoundType.EQ][i] = lb
                continue
            if lb is not None:
                bounds_map[BoundType.LO][i] = lb
            if ub is not None:
                bounds_map[BoundType.UP][i] = ub
    return bounds_map


def parse_types(
    items: Iterable[ItemType], idx_map: Mapping[int, int]
) -> Mapping[int, int]:
    types_map = {}
    for item in items:
        i = idx_map[id(item)]
        if isinstance(item, VarData):
            if item.is_binary():
                types_map[i] = knitro.KN_VARTYPE_BINARY
            elif item.is_integer():
                types_map[i] = knitro.KN_VARTYPE_INTEGER
            elif item.is_continuous():
                types_map[i] = knitro.KN_VARTYPE_CONTINUOUS
            else:
                msg = f"Variable {item.name} has unsupported type."
                raise ValueError(msg)
    return types_map


def api_set_param(param_type: int) -> Callable[..., None]:
    if param_type == knitro.KN_PARAMTYPE_INTEGER:
        return knitro.KN_set_int_param
    elif param_type == knitro.KN_PARAMTYPE_FLOAT:
        return knitro.KN_set_double_param
    elif param_type == knitro.KN_PARAMTYPE_STRING:
        return knitro.KN_set_char_param
    raise DeveloperError(f"Unsupported KNITRO parameter type: {param_type}")


def api_get_values(
    item_type: type[ItemType], value_type: ValueType
) -> Callable[..., list[float] | None]:
    if item_type is VarData:
        if value_type == ValueType.PRIMAL:
            return knitro.KN_get_var_primal_values
        elif value_type == ValueType.DUAL:
            return knitro.KN_get_var_dual_values
    elif item_type is ConstraintData:
        if value_type == ValueType.DUAL:
            return knitro.KN_get_con_dual_values
        elif value_type == ValueType.PRIMAL:
            return knitro.KN_get_con_values
    raise DeveloperError(
        f"Unsupported KNITRO item type or value type: {item_type}, {value_type}"
    )


def api_add_items(item_type: type[ItemType]) -> Callable[..., list[int] | None]:
    if item_type is VarData:
        return knitro.KN_add_vars
    elif item_type is ConstraintData:
        return knitro.KN_add_cons
    raise DeveloperError(f"Unsupported KNITRO item type: {item_type}")


def api_set_bnds(
    item_type: type[ItemType], bound_type: BoundType
) -> Callable[..., None]:
    if item_type is VarData:
        if bound_type == BoundType.EQ:
            return knitro.KN_set_var_fxbnds
        elif bound_type == BoundType.LO:
            return knitro.KN_set_var_lobnds
        elif bound_type == BoundType.UP:
            return knitro.KN_set_var_upbnds
    elif item_type is ConstraintData:
        if bound_type == BoundType.EQ:
            return knitro.KN_set_con_eqbnds
        elif bound_type == BoundType.LO:
            return knitro.KN_set_con_lobnds
        elif bound_type == BoundType.UP:
            return knitro.KN_set_con_upbnds
    raise DeveloperError(
        f"Unsupported KNITRO item type or bound type: {item_type}, {bound_type}"
    )


def api_set_types(item_type: type[ItemType]) -> Callable[..., None]:
    if item_type is VarData:
        return knitro.KN_set_var_types
    raise DeveloperError(f"Unsupported KNITRO item type: {item_type}")


def api_add_struct(is_obj: bool, structure_type: StructureType) -> Callable[..., None]:
    if is_obj:
        if structure_type == StructureType.CONSTANT:
            return knitro.KN_add_obj_constant
        elif structure_type == StructureType.LINEAR:
            return knitro.KN_add_obj_linear_struct
        elif structure_type == StructureType.QUADRATIC:
            return knitro.KN_add_obj_quadratic_struct
    else:
        if structure_type == StructureType.CONSTANT:
            return knitro.KN_add_con_constants
        elif structure_type == StructureType.LINEAR:
            return knitro.KN_add_con_linear_struct
        elif structure_type == StructureType.QUADRATIC:
            return knitro.KN_add_con_quadratic_struct
    raise DeveloperError(
        f"Unsupported KNITRO structure type: is_obj={is_obj}, structure_type={structure_type}"
    )


class Engine:
    """A wrapper around the KNITRO API for a single optimization problem."""

    has_objective: bool
    maps: Mapping[type[ItemData], MutableMapping[int, int]]
    nonlinear_map: MutableMapping[int | None, NonlinearExpressionData]
    nonlinear_diff_order: int

    _kc: Any | None
    _status: int | None

    def __init__(self, *, nonlinear_diff_order: int = 2) -> None:
        self.has_objective = False
        # Maps:
        # VarData -> {id(var): idx in KNITRO}
        # ConstraintData -> {id(con): idx in KNITRO}
        self.maps = MappingProxyType({VarData: {}, ConstraintData: {}})
        # Nonlinear map:
        # None -> objective nonlinear expression
        # idx_con -> constraint nonlinear expression
        self.nonlinear_map = {}
        self.nonlinear_diff_order = nonlinear_diff_order
        self._kc = None
        self._status = None

    def __enter__(self) -> "Engine":
        self.renew()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    def renew(self) -> None:
        self.close()
        self._kc = Package.create_context()
        self.has_objective = False
        for item_type in self.maps:
            self.maps[item_type].clear()
        self.nonlinear_map.clear()
        # TODO: remove this when the tolerance tests are fixed in test_solvers
        tol = 1e-8
        self.set_options(ftol=tol, opttol=tol, xtol=tol)

    def close(self) -> None:
        if self._kc is not None:
            self.execute(knitro.KN_free)
            self._kc = None

    T = TypeVar("T")

    def execute(self, api_func: Callable[..., T], *args, **kwargs) -> T:
        if self._kc is None:
            msg = "KNITRO context has not been initialized or has been freed."
            raise RuntimeError(msg)
        return api_func(self._kc, *args, **kwargs)

    def add_vars(self, variables: Sequence[VarData]) -> None:
        self.add_items(VarData, variables)
        self.set_types(VarData, variables)
        self.set_bounds(VarData, variables)

    def add_cons(self, cons: Sequence[ConstraintData]) -> None:
        self.add_items(ConstraintData, cons)
        self.set_bounds(ConstraintData, cons)
        self.set_con_structures(cons)

    def set_obj(self, obj: ObjectiveData) -> None:
        self.has_objective = True
        self.set_obj_goal(obj.sense)
        self.set_obj_structures(obj)

    def set_options(self, **options) -> None:
        for param, val in options.items():
            self.set_option(param, val)

    def set_outlev(self, level: int | None = None) -> None:
        if level is None:
            level = knitro.KN_OUTLEV_ALL
        self.set_options(outlev=level)

    def set_time_limit(self, time_limit: float) -> None:
        self.set_options(maxtime_cpu=time_limit)

    def set_num_threads(self, nthreads: int) -> None:
        self.set_options(threads=nthreads)

    def solve(self) -> int:
        self.register_callbacks()
        self._status = self.execute(knitro.KN_solve)
        return self._status

    def get_idx_vars(self, variables: Iterable[VarData]) -> list[int]:
        return self.get_idxs(VarData, variables)

    def get_status(self) -> int:
        if self._status is None:
            msg = "Solver has not been run. No status is available!"
            raise RuntimeError(msg)
        return self._status

    def get_number_iters(self) -> int:
        return self.execute(knitro.KN_get_number_iters)

    def get_mip_number_nodes(self) -> int:
        return self.execute(knitro.KN_get_mip_number_nodes)

    def get_mip_abs_gap(self) -> float:
        return self.execute(knitro.KN_get_mip_abs_gap)

    def get_mip_rel_gap(self) -> float:
        return self.execute(knitro.KN_get_mip_rel_gap)

    def get_mip_number_solves(self) -> int:
        return self.execute(knitro.KN_get_mip_number_solves)

    def get_num_solutions(self) -> int:
        _, _, x, _ = self.execute(knitro.KN_get_solution)
        return 1 if x is not None else 0

    def get_solve_time(self) -> float:
        return self.execute(knitro.KN_get_solve_time_real)

    def get_obj_value(self) -> float | None:
        if not self.has_objective:
            return None
        if self.get_solution_status() in {
            SolutionStatus.optimal,
            SolutionStatus.feasible,
        }:
            return self.execute(knitro.KN_get_obj_value)
        return None

    def get_obj_bound(self) -> float | None:
        if not self.has_objective:
            return None
        return self.execute(knitro.KN_get_mip_relaxation_bnd)

    def get_idxs(
        self, item_type: type[ItemType], items: Iterable[ItemType]
    ) -> list[int]:
        idx_map = self.maps[item_type]
        return [idx_map[id(item)] for item in items]

    def get_values(
        self,
        item_type: type[ItemType],
        value_type: ValueType,
        items: Iterable[ItemType],
    ) -> list[float] | None:
        func = api_get_values(item_type, value_type)
        idxs = self.get_idxs(item_type, items)
        return self.execute(func, idxs)

    def set_option(self, param: str, val) -> None:
        param_id = self.execute(knitro.KN_get_param_id, param)
        param_type = self.execute(knitro.KN_get_param_type, param_id)
        func = api_set_param(param_type)
        self.execute(func, param_id, val)

    def set_obj_goal(self, sense: ObjectiveSense) -> None:
        obj_goal = (
            knitro.KN_OBJGOAL_MINIMIZE
            if sense == ObjectiveSense.minimize
            else knitro.KN_OBJGOAL_MAXIMIZE
        )
        self.execute(knitro.KN_set_obj_goal, obj_goal)

    def add_items(self, item_type: type[ItemType], items: Sequence[ItemType]) -> None:
        func = api_add_items(item_type)
        idxs = self.execute(func, len(items))
        if idxs is not None:
            self.maps[item_type].update(zip(map(id, items), idxs))

    def set_bounds(self, item_type: type[ItemType], items: Iterable[ItemType]) -> None:
        bounds_map = parse_bounds(items, self.maps[item_type])
        for bound_type, bounds in bounds_map.items():
            if not bounds:
                continue

            func = api_set_bnds(item_type, bound_type)
            self.execute(func, bounds.keys(), bounds.values())

    def set_types(self, item_type: type[ItemType], items: Iterable[ItemType]) -> None:
        types_map = parse_types(items, self.maps[item_type])
        if types_map:
            func = api_set_types(item_type)
            self.execute(func, types_map.keys(), types_map.values())

    def set_con_structures(self, cons: Iterable[ConstraintData]) -> None:
        for con in cons:
            i = self.maps[ConstraintData][id(con)]
            self.add_structures(i, con.body)

    def set_obj_structures(self, obj: ObjectiveData) -> None:
        self.add_structures(None, obj.expr)

    def add_structures(self, i: int | None, expr) -> None:
        repn = generate_standard_repn(expr)
        if repn is None:
            return

        is_obj = i is None
        base_args = () if is_obj else (i,)
        structure_type_seq: list[StructureType] = []
        args_seq: list[tuple[Any, ...]] = []

        if repn.constant is not None:
            structure_type_seq += [StructureType.CONSTANT]
            args_seq += [(repn.constant,)]

        if repn.linear_vars:
            idx_lin_vars = self.get_idx_vars(repn.linear_vars)
            lin_coefs = list(repn.linear_coefs)
            structure_type_seq += [StructureType.LINEAR]
            args_seq += [(idx_lin_vars, lin_coefs)]

        if repn.quadratic_vars:
            quad_vars1, quad_vars2 = zip(*repn.quadratic_vars)
            idx_quad_vars1 = self.get_idx_vars(quad_vars1)
            idx_quad_vars2 = self.get_idx_vars(quad_vars2)
            quad_coefs = list(repn.quadratic_coefs)
            structure_type_seq += [StructureType.QUADRATIC]
            args_seq += [(idx_quad_vars1, idx_quad_vars2, quad_coefs)]

        for structure_type, args in zip(structure_type_seq, args_seq):
            func = api_add_struct(is_obj, structure_type)
            self.execute(func, *base_args, *args)

        if repn.nonlinear_expr is not None:
            self.nonlinear_map[i] = NonlinearExpressionData(
                repn.nonlinear_expr,
                repn.nonlinear_vars,
                var_map=self.maps[VarData],
                diff_order=self.nonlinear_diff_order,
            )

    def add_callback(
        self, i: int | None, expr: NonlinearExpressionData, callback: Callback
    ) -> None:
        is_obj = i is None
        idx_cons = [i] if not is_obj else None
        cb = self.execute(knitro.KN_add_eval_callback, is_obj, idx_cons, callback.func)

        if expr.diff_order >= 1:
            idx_vars = self.get_idx_vars(expr.grad_vars)
            obj_grad_idx_vars = idx_vars if is_obj else None
            jac_idx_cons = [i] * len(idx_vars) if not is_obj else None
            jac_idx_vars = idx_vars if not is_obj else None
            self.execute(
                knitro.KN_set_cb_grad,
                cb,
                obj_grad_idx_vars,
                jac_idx_cons,
                jac_idx_vars,
                callback.grad,
            )

        if expr.diff_order >= 2:
            hess_vars1, hess_vars2 = zip(*expr.hess_vars)
            hess_idx_vars1 = self.get_idx_vars(hess_vars1)
            hess_idx_vars2 = self.get_idx_vars(hess_vars2)
            self.execute(
                knitro.KN_set_cb_hess, cb, hess_idx_vars1, hess_idx_vars2, callback.hess
            )

    def register_callbacks(self) -> None:
        for i, expr in self.nonlinear_map.items():
            self.register_callback(i, expr)

    def register_callback(self, i: int | None, expr: NonlinearExpressionData) -> None:
        callback = build_callback_handler(expr, idx=i).expand()
        self.add_callback(i, expr, callback)

    def get_solution_status(self) -> SolutionStatus:
        """
        Map KNITRO status codes to Pyomo SolutionStatus values.

        See https://www.artelys.com/app/docs/knitro/3_referenceManual/returnCodes.html
        """
        if self._status is None:
            msg = "Solver has not been run. No solution status is available!"
            raise RuntimeError(msg)
        elif self._status in {
            knitro.KN_RC_OPTIMAL,
            knitro.KN_RC_OPTIMAL_OR_SATISFACTORY,
            knitro.KN_RC_NEAR_OPT,
        }:
            return SolutionStatus.optimal
        elif self._status in {
            knitro.KN_RC_FEAS_XTOL,
            knitro.KN_RC_FEAS_NO_IMPROVE,
            knitro.KN_RC_FEAS_FTOL,
            -103,  # KN_RC_FEAS_BEST
            -104,  # KN_RC_FEAS_MULTISTART
            knitro.KN_RC_ITER_LIMIT_FEAS,
            knitro.KN_RC_TIME_LIMIT_FEAS,
            knitro.KN_RC_FEVAL_LIMIT_FEAS,
            knitro.KN_RC_MIP_EXH_FEAS,
            knitro.KN_RC_MIP_TERM_FEAS,
            knitro.KN_RC_MIP_SOLVE_LIMIT_FEAS,
            knitro.KN_RC_MIP_NODE_LIMIT_FEAS,
        }:
            return SolutionStatus.feasible
        elif self._status in {
            knitro.KN_RC_INFEASIBLE,
            knitro.KN_RC_INFEAS_CON_BOUNDS,
            knitro.KN_RC_INFEAS_VAR_BOUNDS,
            knitro.KN_RC_INFEAS_MULTISTART,
        }:
            return SolutionStatus.infeasible
        else:
            return SolutionStatus.noSolution

    def get_termination_condition(self) -> TerminationCondition:
        """
        Map KNITRO status codes to Pyomo TerminationCondition values.

        See https://www.artelys.com/app/docs/knitro/3_referenceManual/returnCodes.html
        """
        if self._status is None:
            msg = "Solver has not been run. No termination condition is available!"
            raise RuntimeError(msg)
        elif self._status in {
            knitro.KN_RC_OPTIMAL,
            knitro.KN_RC_OPTIMAL_OR_SATISFACTORY,
            knitro.KN_RC_NEAR_OPT,
        }:
            return TerminationCondition.convergenceCriteriaSatisfied
        elif self._status in {
            knitro.KN_RC_INFEAS_NO_IMPROVE,
            knitro.KN_RC_INFEAS_MULTISTART,
        }:
            return TerminationCondition.locallyInfeasible
        elif self._status in {
            knitro.KN_RC_INFEASIBLE,
            knitro.KN_RC_INFEAS_CON_BOUNDS,
            knitro.KN_RC_INFEAS_VAR_BOUNDS,
        }:
            return TerminationCondition.provenInfeasible
        elif self._status in {knitro.KN_RC_UNBOUNDED, knitro.KN_RC_UNBOUNDED_OR_INFEAS}:
            return TerminationCondition.infeasibleOrUnbounded
        elif self._status in {
            knitro.KN_RC_ITER_LIMIT_FEAS,
            knitro.KN_RC_FEVAL_LIMIT_FEAS,
            knitro.KN_RC_MIP_EXH_FEAS,
            knitro.KN_RC_MIP_TERM_FEAS,
            knitro.KN_RC_MIP_SOLVE_LIMIT_FEAS,
            knitro.KN_RC_MIP_NODE_LIMIT_FEAS,
            knitro.KN_RC_ITER_LIMIT_INFEAS,
            knitro.KN_RC_FEVAL_LIMIT_INFEAS,
            knitro.KN_RC_MIP_EXH_INFEAS,
            knitro.KN_RC_MIP_SOLVE_LIMIT_INFEAS,
            knitro.KN_RC_MIP_NODE_LIMIT_INFEAS,
        }:
            return TerminationCondition.iterationLimit
        elif self._status in {
            knitro.KN_RC_TIME_LIMIT_FEAS,
            knitro.KN_RC_TIME_LIMIT_INFEAS,
        }:
            return TerminationCondition.maxTimeLimit
        elif self._status == knitro.KN_RC_USER_TERMINATION:
            return TerminationCondition.interrupted
        elif -500 >= self._status >= -600:
            return TerminationCondition.error
        else:
            return TerminationCondition.unknown



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import os
from os.path import join
from filecmp import cmp

import pyomo.common.unittest as unittest

import pyomo.common
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

from pyomo.core import ConcreteModel
from pyomo.opt import ResultsFormat, SolverResults, SolverFactory

currdir = this_file_dir()
deleteFiles = True

old_ignore_time = None


def setUpModule():
    global old_ignore_time
    old_ignore_time = SolverResults.default_print_options.ignore_time
    SolverResults.default_print_options.ignore_time = True


def tearDownModule():
    SolverResults.default_print_options.ignore_time = old_ignore_time


cplexamp_available = False


class mock_all(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global cplexamp_available
        import pyomo.environ
        from pyomo.solvers.tests.solvers import test_solver_cases as _test_solver_cases

        cplexamp_available = _test_solver_cases('cplex', 'nl').available

    def setUp(self):
        self.do_setup(False)

    def do_setup(self, flag):
        TempfileManager.push()
        if flag:
            if not cplexamp_available:
                self.skipTest("The 'cplexamp' command is not available")
            self.asl = SolverFactory('asl:cplexamp')
        else:
            self.asl = SolverFactory('_mock_asl:cplexamp')

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())
        self.asl = None

    def test_path(self):
        """Verify that the ASL path is what is expected"""
        if type(self.asl) == 'ASL':
            self.assertEqual(
                self.asl.executable.split(os.sep)[-1],
                "ASL" + pyomo.common.executable_extension,
            )

    def test_solve4(self):
        """Test ASL - test4.nl"""
        _log = TempfileManager.create_tempfile(".test_solve4.log")
        _out = TempfileManager.create_tempfile(".test_solve4.txt")
        TempfileManager.add_tempfile(join(currdir, 'test4.soln'), exists=False)

        results = self.asl.solve(
            join(currdir, "test4.nl"), logfile=_log, suffixes=['.*']
        )
        results.write(filename=_out, times=False, format='json')
        _baseline = join(currdir, "test4_asl.txt")
        with open(_out, 'r') as out, open(_baseline, 'r') as txt:
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )

    #
    # This test is disabled, but it's useful for interactively exercising
    # the option specifications of a solver
    #
    def Xtest_options(self):
        """Test ASL options behavior"""
        results = self.asl.solve(
            currdir + "bell3a.mps",
            logfile=currdir + "test_options.log",
            options="sec=0.1 foo=1 bar='a=b c=d' xx_zz=yy",
            suffixes=['.*'],
        )
        results.write(filename=currdir + "test_options.txt", times=False)
        _out, _log = join(currdir, "test_options.txt"), join(currdir, "test4_asl.txt")
        self.assertTrue(cmp(_out, _log), msg="Files %s and %s differ" % (_out, _log))
        # os.remove(currdir+"test4.sol")
        # os.remove(currdir+"test_solve4.log")

    def test_error1(self):
        """Bad results format"""
        try:
            model = ConcreteModel()
            results = self.asl.solve(model, format=ResultsFormat.sol, suffixes=['.*'])
            self.fail("test_error1")
        except ValueError:
            pass

    def test_error2(self):
        """Bad solve option"""
        try:
            model = ConcreteModel()
            results = self.asl.solve(model, foo="bar")
            self.fail("test_error2")
        except ValueError:
            pass

    def test_error3(self):
        """Bad solve option"""
        try:
            results = self.asl.solve(currdir + "model.py", foo="bar")
            self.fail("test_error3")
        except ValueError:
            pass


class mip_all(mock_all):
    def setUp(self):
        self.do_setup(True)


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"This is the deprecated pyomo.core.expr.current module"

import enum
import math

#
# Common intrinsic functions
#
import pyomo.core.expr.expr_common as common
from pyomo.core.expr.expr_common import clone_counter, _mode

from pyomo.core.expr import (
    Mode,
    # from pyomo.core.expr.base
    ExpressionBase,
    # pyomo.core.expr.visitor
    evaluate_expression,
    expression_to_string,
    polynomial_degree,
    clone_expression,
    sizeof_expression,
    # pyomo.core.expr.numeric_expr
    NumericExpression,
    NumericValue,
    native_types,
    nonpyomo_leaf_types,
    native_numeric_types,
    value,
    nonlinear_expression,
    linear_expression,
    NegationExpression,
    NPV_NegationExpression,
    ExternalFunctionExpression,
    NPV_ExternalFunctionExpression,
    PowExpression,
    NPV_PowExpression,
    ProductExpression,
    NPV_ProductExpression,
    MonomialTermExpression,
    DivisionExpression,
    NPV_DivisionExpression,
    SumExpressionBase,
    NPV_SumExpression,
    SumExpression,
    Expr_ifExpression,
    NPV_Expr_ifExpression,
    UnaryFunctionExpression,
    NPV_UnaryFunctionExpression,
    AbsExpression,
    NPV_AbsExpression,
    LinearExpression,
    decompose_term,
    LinearDecompositionError,
    NPV_expression_types,
    Expr_if,
    ceil,
    floor,
    exp,
    log,
    log10,
    sqrt,
    sin,
    cos,
    tan,
    sinh,
    cosh,
    tanh,
    asin,
    acos,
    atan,
    asinh,
    acosh,
    atanh,
    # pyomo.core.expr.numvalue
    as_numeric,
    # pyomo.core.expr.logical_expr
    native_logical_types,
    BooleanValue,
    BooleanConstant,
    BooleanExpressionBase,
    lnot,
    equivalent,
    xor,
    implies,
    land,
    lor,
    exactly,
    atmost,
    atleast,
    UnaryBooleanExpression,
    NotExpression,
    BinaryBooleanExpression,
    EquivalenceExpression,
    XorExpression,
    ImplicationExpression,
    NaryBooleanExpression,
    AndExpression,
    OrExpression,
    ExactlyExpression,
    AtMostExpression,
    AtLeastExpression,
    special_boolean_atom_types,
    # pyomo.core.expr.relational_expr
    RelationalExpression,
    RangedExpression,
    InequalityExpression,
    EqualityExpression,
    inequality,
    # pyomo.core.expr.template_expr
    TemplateExpressionError,
    GetItemExpression,
    Numeric_GetItemExpression,
    Boolean_GetItemExpression,
    Structural_GetItemExpression,
    NPV_Numeric_GetItemExpression,
    NPV_Boolean_GetItemExpression,
    NPV_Structural_GetItemExpression,
    GetAttrExpression,
    Numeric_GetAttrExpression,
    Boolean_GetAttrExpression,
    Structural_GetAttrExpression,
    NPV_Numeric_GetAttrExpression,
    NPV_Boolean_GetAttrExpression,
    NPV_Structural_GetAttrExpression,
    CallExpression,
    TemplateSumExpression,
    IndexTemplate,
    resolve_template,
    ReplaceTemplateExpression,
    substitute_template_expression,
    substitute_getitem_with_param,
    substitute_template_with_value,
    templatize_rule,
    templatize_constraint,
    # pyomo.core.expr.visitor
    SymbolMap,
    StreamBasedExpressionVisitor,
    SimpleExpressionVisitor,
    ExpressionValueVisitor,
    replace_expressions,
    ExpressionReplacementVisitor,
    FixedExpressionError,
    NonConstantExpressionError,
    identify_components,
    identify_variables,
    identify_mutable_parameters,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import attempt_import
from pyomo.common.deprecation import relocated_module_attribute
from pyomo.scripting.pyomo_parser import add_subparser

relocated_module_attribute(
    'QtApp', 'pyomo.contrib.viewer.pyomo_qtapp.QtApp', version='6.9.3'
)

qtconsole_app, qtconsole_available = attempt_import(
    "qtconsole.qtconsoleapp", defer_import=False
)


def main(*args):
    # Import the Qt infrastructure (if it exists)
    import pyomo.contrib.viewer.qt as myqt

    if not myqt.available or not qtconsole_available:
        errors = list(myqt.import_errors)
        if not qtconsole_available:
            errors.append(qtconsole_app._moduleunavailable_message())
        print("qt not available\n    " + "\n    ".join(errors))
        return

    # Ensure that all Pyomo plugins have been registered
    import pyomo.environ

    # Import & run the Qt application
    from pyomo.contrib.viewer.pyomo_qtapp import QtApp

    QtApp.launch_instance()


# Add a subparser for the model-viewer command
add_subparser(
    "model-viewer",
    func=main,
    help="Run the Pyomo model viewer",
    add_help=False,
    description="This runs the Pyomo model viewer",
)

if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

from pyomo.common.deprecation import relocated_module_attribute


class Bar:
    data = 42


def __getattr__(name):
    if name.startswith('Foo'):
        return name[3:]
    raise AttributeError("module '%s' has no attribute '%s'" % (__name__, name))


relocated_module_attribute('Foo', 'pyomo.common.tests.test_deprecated.Bar', 'test')
relocated_module_attribute('Foo_2', 'pyomo.common.tests.relocated.Bar', 'test')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import pyomo.common.unittest as unittest

import pyomo.environ as pyo
from pyomo.core.expr.compare import compare_expressions
from pyomo.contrib.mpc.modeling.terminal import (
    get_penalty_at_time,
    get_terminal_penalty,
)
from pyomo.contrib.mpc.data.scalar_data import ScalarData


class TestTerminalPenalty(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=list(range(n_time_points)))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time, m.comp, initialize={(i, j): 1.1 * i for i, j in m.time * m.comp}
        )
        m.input = pyo.Var(m.time, initialize={i: 3.3 * i for i in m.time})
        return m

    def test_get_penalty(self):
        m = self._make_model()

        variables = [pyo.Reference(m.var[:, "A"]), m.input]
        target = ScalarData({m.var[:, "A"]: 4.4, m.input[:]: 5.5})
        weight_data = {
            pyo.ComponentUID(m.var[:, "A"]): 10.0,
            pyo.ComponentUID(m.input[:]): 0.2,
        }
        tp = 1
        m.var_set, m.penalty = get_penalty_at_time(
            variables, tp, target, weight_data=weight_data
        )
        for i in m.var_set:
            pred_expr = (
                10.0 * (m.var[tp, "A"] - 4.4) ** 2
                if i == 0
                else 0.2 * (m.input[tp] - 5.5) ** 2
            )
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i].expr))
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i].expr))

    def test_get_terminal_penalty(self):
        m = self._make_model()

        variables = [pyo.Reference(m.var[:, "A"]), m.input]
        target = ScalarData({m.var[:, "A"]: 4.4, m.input[:]: 5.5})
        weight_data = {
            pyo.ComponentUID(m.var[:, "A"]): 10.0,
            pyo.ComponentUID(m.input[:]): 0.2,
        }
        m.var_set, m.penalty = get_terminal_penalty(
            variables, m.time, target, weight_data=weight_data
        )

        for i in m.var_set:
            tf = m.time.last()
            pred_expr = (
                10.0 * (m.var[tf, "A"] - 4.4) ** 2
                if i == 0
                else 0.2 * (m.input[tf] - 5.5) ** 2
            )
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i].expr))
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i].expr))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Tests deactivation of trivial constraints."""

import pyomo.common.unittest as unittest
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.environ import Constraint, ConcreteModel, TransformationFactory, Var


class TestTrivialConstraintDeactivator(unittest.TestCase):
    """Tests deactivation of trivial constraints."""

    def test_deactivate_trivial_constraints(self):
        """Test for deactivation of trivial constraints."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.c = Constraint(expr=m.v1 <= m.v2)
        m.c2 = Constraint(expr=m.v2 >= m.v3)
        m.c3 = Constraint(expr=m.v1 <= 5)
        m.v1.fix()

        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(m)
        self.assertTrue(m.c.active)
        self.assertTrue(m.c2.active)
        self.assertFalse(m.c3.active)

    def test_deactivate_trivial_constraints_return_list(self):
        """Test for deactivation of trivial constraints."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.c = Constraint(expr=m.v1 <= m.v2)
        m.c2 = Constraint(expr=m.v2 >= m.v3)
        m.c3 = Constraint(expr=m.v1 <= 5)
        m.v1.fix()

        trivial = []
        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(
            m, return_trivial=trivial
        )
        self.assertTrue(m.c.active)
        self.assertTrue(m.c2.active)
        self.assertFalse(m.c3.active)
        self.assertEqual(len(trivial), 1)
        self.assertIs(trivial[0], m.c3)

    def test_deactivate_trivial_constraints_revert(self):
        """Test for reversion of trivial constraint deactivation."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.c = Constraint(expr=m.v1 <= m.v2)
        m.c2 = Constraint(expr=m.v2 >= m.v3)
        m.c3 = Constraint(expr=m.v1 <= 5)
        m.v1.fix()

        xfrm = TransformationFactory('contrib.deactivate_trivial_constraints')
        xfrm.apply_to(m, tmp=True)
        self.assertTrue(m.c.active)
        self.assertTrue(m.c2.active)
        self.assertFalse(m.c3.active)

        xfrm.revert(m)
        self.assertTrue(m.c3.active)

    def test_trivial_constraints_lb_conflict(self):
        """Test for violated trivial constraint lower bound."""
        with self.assertRaisesRegex(
            InfeasibleConstraintException,
            "Trivial constraint c violates LB 2.0 ≤ BODY 1.",
        ):
            self._trivial_constraints_lb_conflict()

    def _trivial_constraints_lb_conflict(self):
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.c = Constraint(expr=m.v1 >= 2)
        m.v1.fix()
        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(m)

    def test_trivial_constraints_ub_conflict(self):
        """Test for violated trivial constraint upper bound."""
        with self.assertRaisesRegex(
            InfeasibleConstraintException,
            "Trivial constraint c violates BODY 1 ≤ UB 0.0.",
        ):
            self._trivial_constraints_ub_conflict()

    def _trivial_constraints_ub_conflict(self):
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.c = Constraint(expr=m.v1 <= 0)
        m.v1.fix()
        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(m)

    def test_trivial_constraint_due_to_0_coefficient(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.y.fix(0)
        m.c = Constraint(expr=m.x * m.y >= 0)

        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(m)

        self.assertFalse(m.c.active)

    def test_higher_degree_trivial_constraint(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()
        m.c = Constraint(expr=(m.x**2 + m.y) * m.z >= -8)
        m.z.fix(0)
        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(m)
        self.assertFalse(m.c.active)

    def test_trivial_linear_constraint_due_to_cancellation(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=m.x - m.x <= 0)

        TransformationFactory('contrib.deactivate_trivial_constraints').apply_to(m)

        self.assertFalse(m.c.active)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

from io import StringIO
import sys
import logging

import pyomo.common.unittest as unittest

from pyomo.contrib.trustregion.util import IterationLogger, minIgnoreNone, maxIgnoreNone
from pyomo.common.log import LoggingIntercept


class TestUtil(unittest.TestCase):
    def setUp(self):
        self.iterLogger = IterationLogger()
        self.iteration = 0
        self.thetak = 10.0
        self.objk = 5.0
        self.radius = 1.0
        self.stepNorm = 0.25

    def tearDown(self):
        pass

    def test_minIgnoreNone(self):
        a = 1
        b = 2
        self.assertEqual(minIgnoreNone(a, b), a)
        a = None
        self.assertEqual(minIgnoreNone(a, b), b)
        a = 1
        b = None
        self.assertEqual(minIgnoreNone(a, b), a)
        a = None
        self.assertEqual(minIgnoreNone(a, b), None)

    def test_maxIgnoreNone(self):
        a = 1
        b = 2
        self.assertEqual(maxIgnoreNone(a, b), b)
        a = None
        self.assertEqual(maxIgnoreNone(a, b), b)
        a = 1
        b = None
        self.assertEqual(maxIgnoreNone(a, b), a)
        a = None
        self.assertEqual(maxIgnoreNone(a, b), None)

    def test_IterationRecord(self):
        self.iterLogger.newIteration(
            self.iteration, self.thetak, self.objk, self.radius, self.stepNorm
        )
        self.assertEqual(len(self.iterLogger.iterations), 1)
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, 5.0)

    def test_logIteration(self):
        self.iterLogger.newIteration(
            self.iteration, self.thetak, self.objk, self.radius, self.stepNorm
        )
        OUTPUT = StringIO()
        with LoggingIntercept(OUTPUT, 'pyomo.contrib.trustregion', logging.INFO):
            self.iterLogger.logIteration()
        self.assertIn('Iteration 0', OUTPUT.getvalue())
        self.assertIn('feasibility =', OUTPUT.getvalue())
        self.assertIn('stepNorm =', OUTPUT.getvalue())

    def test_updateIteration(self):
        self.iterLogger.newIteration(
            self.iteration, self.thetak, self.objk, self.radius, self.stepNorm
        )
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, self.objk)
        self.assertEqual(self.iterLogger.iterations[0].feasibility, self.thetak)
        self.assertEqual(self.iterLogger.iterations[0].trustRadius, self.radius)
        self.assertEqual(self.iterLogger.iterations[0].stepNorm, self.stepNorm)
        self.iterLogger.updateIteration(feasibility=5.0)
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, self.objk)
        self.assertEqual(self.iterLogger.iterations[0].feasibility, 5.0)
        self.assertEqual(self.iterLogger.iterations[0].trustRadius, self.radius)
        self.assertEqual(self.iterLogger.iterations[0].stepNorm, self.stepNorm)
        self.iterLogger.updateIteration(objectiveValue=0.1)
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, 0.1)
        self.assertEqual(self.iterLogger.iterations[0].feasibility, 5.0)
        self.assertEqual(self.iterLogger.iterations[0].trustRadius, self.radius)
        self.assertEqual(self.iterLogger.iterations[0].stepNorm, self.stepNorm)
        self.iterLogger.updateIteration(trustRadius=100)
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, 0.1)
        self.assertEqual(self.iterLogger.iterations[0].feasibility, 5.0)
        self.assertEqual(self.iterLogger.iterations[0].trustRadius, 100)
        self.assertEqual(self.iterLogger.iterations[0].stepNorm, self.stepNorm)
        self.iterLogger.updateIteration(stepNorm=1)
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, 0.1)
        self.assertEqual(self.iterLogger.iterations[0].feasibility, 5.0)
        self.assertEqual(self.iterLogger.iterations[0].trustRadius, 100)
        self.assertEqual(self.iterLogger.iterations[0].stepNorm, 1)
        self.iterLogger.updateIteration(
            feasibility=10.0, objectiveValue=0.2, trustRadius=1000, stepNorm=10
        )
        self.assertEqual(self.iterLogger.iterations[0].objectiveValue, 0.2)
        self.assertEqual(self.iterLogger.iterations[0].feasibility, 10.0)
        self.assertEqual(self.iterLogger.iterations[0].trustRadius, 1000)
        self.assertEqual(self.iterLogger.iterations[0].stepNorm, 10)

    def test_printIteration(self):
        self.iterLogger.newIteration(
            self.iteration, self.thetak, self.objk, self.radius, self.stepNorm
        )
        OUTPUT = StringIO()
        sys.stdout = OUTPUT
        self.iterLogger.printIteration()
        sys.stdout = sys.__stdout__
        self.assertIn(str(self.radius), OUTPUT.getvalue())
        self.assertIn(str(self.iteration), OUTPUT.getvalue())
        self.assertIn(str(self.thetak), OUTPUT.getvalue())
        self.assertIn(str(self.objk), OUTPUT.getvalue())
        self.assertIn(str(self.stepNorm), OUTPUT.getvalue())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import re
import sys
import time
import logging
import shlex
from typing import overload

from pyomo.common import Factory
from pyomo.common.enums import SolverAPIVersion
from pyomo.common.errors import ApplicationError
from pyomo.common.collections import Bunch

from pyomo.opt.base.convert import convert_problem
from pyomo.opt.base.formats import ResultsFormat
import pyomo.opt.base.results

logger = logging.getLogger('pyomo.opt')


# The version string is first searched for trunk/Trunk, and if
# found a tuple of infinities is returned. Otherwise, the first
# match of number[.number] where [.number] can repeat 1-3 times
# is used, which is translated into a tuple of size matching
# the keyword length (appending 0's when necessary). If no match
# is found None is returned (although one could argue a tuple of
# 0's might be appropriated).
def _extract_version(x, length=4):
    """
    Attempts to extract solver version information from a string.
    """
    assert (1 <= length) and (length <= 4)
    m = re.search('[t,T]runk', x)
    if m is not None:
        # Since most version checks are comparing if the current
        # version is greater/less than some other version, it makes
        # since that a solver advertising trunk should always be greater
        # than a version check, hence returning a tuple of infinities
        return tuple(float('inf') for i in range(length))
    m = re.search(r'[0-9]+(\.[0-9]+){1,3}', x)
    if not m is None:
        version = tuple(int(i) for i in m.group(0).split('.')[:length])
        while len(version) < length:
            version += (0,)
        return version
    return None  # (0,0,0,0)[:length]


class UnknownSolver:
    def __init__(self, *args, **kwds):
        # super(UnknownSolver,self).__init__(**kwds)

        #
        # The 'type' is the class type of the solver instance
        #
        if "type" in kwds:
            self.type = kwds["type"]
        else:  # pragma:nocover
            raise ValueError("Expected option 'type' for UnknownSolver constructor")

        self.options = {}
        self._args = args
        self._kwds = kwds

    #
    # Support "with" statements. Forgetting to call deactivate
    # on Plugins is a common source of memory leaks
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    @classmethod
    def api_version(self):
        """
        Return the public API supported by this interface.

        Returns
        -------
        ~pyomo.common.enums.SolverAPIVersion
            A solver API enum object
        """
        return SolverAPIVersion.V1

    def available(self, exception_flag=True):
        """Determine if this optimizer is available."""
        if exception_flag:
            raise ApplicationError("Solver (%s) not available" % str(self.name))
        return False

    def license_is_valid(self):
        "True if the solver is present and has a valid license (if applicable)"
        return False

    def warm_start_capable(self):
        """True is the solver can accept a warm-start solution."""
        return False

    def solve(self, *args, **kwds):
        """Perform optimization and return an SolverResults object."""
        self._solver_error('solve')

    def reset(self):
        """Reset the state of an optimizer"""
        self._solver_error('reset')

    def set_options(self, istr):
        """Set the options in the optimizer from a string."""
        self._solver_error('set_options')

    def __bool__(self):
        return self.available()

    def __getattr__(self, attr):
        self._solver_error(attr)

    def _solver_error(self, method_name):
        raise RuntimeError(
            """Attempting to use an unavailable solver.

The SolverFactory was unable to create the solver "%s"
and returned an UnknownSolver object.  This error is raised at the point
where the UnknownSolver object was used as if it were valid (by calling
method "%s").

The original solver was created with the following parameters:
\t""" % (self.type, method_name)
            + "\n\t".join("%s: %s" % i for i in sorted(self._kwds.items()))
            + "\n\t_args: %s" % (self._args,)
            + "\n\toptions: %s" % (self.options,)
        )


class SolverFactoryClass(Factory):
    @overload
    def __call__(self, _name: None = None, **kwds) -> "SolverFactoryClass": ...
    @overload
    def __call__(self, _name, **kwds) -> "OptSolver": ...

    def __call__(self, _name=None, **kwds):
        if _name is None:
            return self
        _name = str(_name)
        if ':' in _name:
            _name, subsolver = _name.split(':', 1)
            kwds['solver'] = subsolver
        elif 'solver' in kwds:
            subsolver = kwds['solver']
        else:
            subsolver = None
        opt = None
        try:
            if _name in self._cls:
                opt = self._cls[_name](**kwds)
            else:
                mode = kwds.get('solver_io', 'nl')
                if mode is None:
                    mode = 'nl'
                _implicit_solvers = {'nl': 'asl'}
                if "executable" not in kwds:
                    kwds["executable"] = _name
                if mode in _implicit_solvers:
                    if _implicit_solvers[mode] not in self._cls:
                        raise RuntimeError(
                            "  The solver plugin was not registered.\n"
                            "  Please confirm that the 'pyomo.environ' package has been imported."
                        )
                    opt = self._cls[_implicit_solvers[mode]](**kwds)
                    if opt is not None:
                        opt.set_options('solver=' + _name)
        except:
            err = sys.exc_info()
            logger.warning(
                "Failed to create solver with name '%s':\n%s" % (_name, err[1]),
                exc_info=err,
            )
            opt = None
        if opt is not None and _name != "py" and subsolver is not None:
            # py just creates instance of its subsolver, no need for this option
            opt.set_options('solver=' + subsolver)
        if opt is None:
            opt = UnknownSolver(type=_name, **kwds)
            opt.name = _name
        return opt


#: Global registry/factory for "v1" solver interfaces.
LegacySolverFactory: SolverFactoryClass = SolverFactoryClass('solver type')

SolverFactory = SolverFactoryClass('solver type')
SolverFactory._cls = LegacySolverFactory._cls
SolverFactory._doc = LegacySolverFactory._doc


#
# TODO: It is impossible to load CBC with NL file-io using this function,
#       i.e., SolverFactory("cbc", solver_io='nl'),
#       this is NOT asl:cbc (same with PICO)
# WEH:  Why is there a distinction between SolverFactory('asl:cbc') and SolverFactory('cbc', solver_io='nl')???   This is bad.
#
def check_available_solvers(*args):
    from pyomo.solvers.plugins.solvers.GUROBI import GUROBISHELL
    from pyomo.solvers.plugins.solvers.BARON import BARONSHELL
    from pyomo.solvers.plugins.solvers.mosek_direct import MOSEKDirect

    logging.disable(logging.WARNING)

    ans = []
    for arg in args:
        if not isinstance(arg, tuple):
            name = arg
            arg = (arg,)
        else:
            name = arg[0]
        opt = SolverFactory(*arg)
        if opt is None or isinstance(opt, UnknownSolver):
            continue  # not available

        if not opt.available(exception_flag=False):
            continue  # not available

        if hasattr(opt, 'executable') and opt.executable() is None:
            continue  # not available

        if not opt.license_is_valid():
            continue  # not available

        # At this point, the solver is available (and licensed)
        ans.append(name)

    logging.disable(logging.NOTSET)

    return ans


def _raise_ephemeral_error(name, keyword=""):
    raise AttributeError(
        "The property '%s' can no longer be set directly on "
        "the solver object. It should instead be passed as a "
        "keyword into the solve method%s. It will automatically "
        "be reset to its default value after each invocation of "
        "solve." % (name, keyword)
    )


class OptSolver:
    """A generic optimization solver"""

    #
    # Support "with" statements. Forgetting to call deactivate
    # on Plugins is a common source of memory leaks
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    @classmethod
    def api_version(self):
        """
        Return the public API supported by this interface.

        Returns
        -------
        ~pyomo.common.enums.SolverAPIVersion
            A solver API enum object
        """
        return SolverAPIVersion.V1

    #
    # Adding to help track down invalid code after making
    # the following attributes private
    #
    @property
    def tee(self):
        _raise_ephemeral_error('tee')

    @tee.setter
    def tee(self, val):
        _raise_ephemeral_error('tee')

    @property
    def suffixes(self):
        _raise_ephemeral_error('suffixes')

    @suffixes.setter
    def suffixes(self, val):
        _raise_ephemeral_error('suffixes')

    @property
    def keepfiles(self):
        _raise_ephemeral_error('keepfiles')

    @keepfiles.setter
    def keepfiles(self, val):
        _raise_ephemeral_error('keepfiles')

    @property
    def soln_file(self):
        _raise_ephemeral_error('soln_file')

    @soln_file.setter
    def soln_file(self, val):
        _raise_ephemeral_error('soln_file')

    @property
    def log_file(self):
        _raise_ephemeral_error('log_file')

    @log_file.setter
    def log_file(self, val):
        _raise_ephemeral_error('log_file')

    @property
    def symbolic_solver_labels(self):
        _raise_ephemeral_error('symbolic_solver_labels')

    @symbolic_solver_labels.setter
    def symbolic_solver_labels(self, val):
        _raise_ephemeral_error('symbolic_solver_labels')

    @property
    def warm_start_solve(self):
        _raise_ephemeral_error('warm_start_solve', keyword=" (warmstart)")

    @warm_start_solve.setter
    def warm_start_solve(self, val):
        _raise_ephemeral_error('warm_start_solve', keyword=" (warmstart)")

    @property
    def warm_start_file_name(self):
        _raise_ephemeral_error('warm_start_file_name', keyword=" (warmstart_file)")

    @warm_start_file_name.setter
    def warm_start_file_name(self, val):
        _raise_ephemeral_error('warm_start_file_name', keyword=" (warmstart_file)")

    def __init__(self, **kwds):
        """Constructor"""
        #
        # The 'type' is the class type of the solver instance
        #
        if "type" in kwds:
            self.type = kwds["type"]
        else:  # pragma:nocover
            raise ValueError("Expected option 'type' for OptSolver constructor")

        #
        # The 'name' is either the class type of the solver instance, or a
        # assigned name.
        #
        if "name" in kwds:
            self.name = kwds["name"]
        else:
            self.name = self.type

        if "doc" in kwds:
            self._doc = kwds["doc"]
        else:
            if self.type is None:  # pragma:nocover
                self._doc = ""
            elif self.name == self.type:
                self._doc = "%s OptSolver" % self.name
            else:
                self._doc = "%s OptSolver (type %s)" % (self.name, self.type)
        #
        # Options are persistent, meaning users must modify the
        # options dict directly rather than pass them into _presolve
        # through the solve command. Everything else is reset inside
        # presolve
        #
        self.options = Bunch()
        if 'options' in kwds and not kwds['options'] is None:
            for key in kwds['options']:
                setattr(self.options, key, kwds['options'][key])

        # the symbol map is an attribute of the solver plugin only
        # because it is generated in presolve and used to tag results
        # so they are interpretable - basically, it persists across
        # multiple methods.
        self._smap_id = None

        # These are ephemeral options that can be set by the user during
        # the call to solve, but will be reset to defaults if not given
        self._load_solutions = True
        self._select_index = 0
        self._report_timing = False
        self._suffixes = []
        self._log_file = None
        self._soln_file = None

        # overridden by a solver plugin when it returns sparse results
        self._default_variable_value = None
        # overridden by a solver plugin when it is always available
        self._assert_available = False
        # overridden by a solver plugin to indicate its input file format
        self._problem_format = None
        self._valid_problem_formats = []
        # overridden by a solver plugin to indicate its results file format
        self._results_format = None
        self._valid_result_formats = {}
        self._results_reader = None
        self._problem = None
        self._problem_files = None

        #
        # Used to document meta solvers
        #
        self._metasolver = False

        self._version = None
        #
        # Data for solver callbacks
        #
        self._allow_callbacks = False
        self._callback = {}

        # We define no capabilities for the generic solver; base
        # classes must override this
        self._capabilities = Bunch()

    @staticmethod
    def _options_string_to_dict(istr):
        ans = {}
        istr = istr.strip()
        if not istr:
            return ans
        if istr[0] == "'" or istr[0] == '"':
            istr = eval(istr)
        tokens = shlex.split(istr)
        for token in tokens:
            index = token.find('=')
            if index == -1:
                raise ValueError(
                    "Solver options must have the form option=value: '%s'" % istr
                )
            try:
                val = eval(token[(index + 1) :])
            except:
                val = token[(index + 1) :]
            ans[token[:index]] = val
        return ans

    def default_variable_value(self):
        return self._default_variable_value

    def __bool__(self):
        return self.available()

    def version(self):
        """
        Returns a 4-tuple describing the solver executable version.
        """
        if self._version is None:
            self._version = self._get_version()
        return self._version

    def _get_version(self):
        return None

    def problem_format(self):
        """
        Returns the current problem format.
        """
        return self._problem_format

    def set_problem_format(self, format):
        """
        Set the current problem format (if it's valid) and update
        the results format to something valid for this problem format.
        """
        if format in self._valid_problem_formats:
            self._problem_format = format
        else:
            raise ValueError(
                "%s is not a valid problem format for solver plugin %s" % (format, self)
            )
        self._results_format = self._default_results_format(self._problem_format)

    def results_format(self):
        """
        Returns the current results format.
        """
        return self._results_format

    def set_results_format(self, format):
        """
        Set the current results format (if it's valid for the current
        problem format).
        """
        if (self._problem_format in self._valid_result_formats) and (
            format in self._valid_result_formats[self._problem_format]
        ):
            self._results_format = format
        else:
            raise ValueError(
                "%s is not a valid results format for "
                "problem format %s with solver plugin %s"
                % (format, self._problem_format, self)
            )

    def has_capability(self, cap):
        """
        Returns a boolean value representing whether a solver supports
        a specific feature. Defaults to 'False' if the solver is unaware
        of an option. Expects a string.

        Example:
        # prints True if solver supports sos1 constraints, and False otherwise
        print(solver.has_capability('sos1')

        # prints True is solver supports 'feature', and False otherwise
        print(solver.has_capability('feature')

        Parameters
        ----------
        cap: str
            The feature

        Returns
        -------
        val: bool
            Whether or not the solver has the specified capability.
        """
        if not isinstance(cap, str):
            raise TypeError(
                "Expected argument to be of type '%s', not "
                "'%s'." % (type(str()), type(cap))
            )
        else:
            val = self._capabilities[str(cap)]
            if val is None:
                return False
            else:
                return val

    def available(self, exception_flag=True):
        """True if the solver is available"""
        return True

    def license_is_valid(self):
        "True if the solver is present and has a valid license (if applicable)"
        return True

    def warm_start_capable(self):
        """True is the solver can accept a warm-start solution"""
        return False

    def solve(self, *args, **kwds):
        """Solve the problem"""

        self.available(exception_flag=True)
        #
        # If the inputs are models, then validate that they have been
        # constructed! Collect suffix names to try and import from solution.
        #
        from pyomo.core.base.block import BlockData
        import pyomo.core.base.suffix
        from pyomo.core.kernel.block import IBlock
        import pyomo.core.kernel.suffix

        _model = None
        for arg in args:
            if isinstance(arg, (BlockData, IBlock)):
                if isinstance(arg, BlockData):
                    if not arg.is_constructed():
                        raise RuntimeError(
                            "Attempting to solve model=%s with unconstructed "
                            "component(s)" % (arg.name,)
                        )

                _model = arg
                # import suffixes must be on the top-level model
                if isinstance(arg, BlockData):
                    model_suffixes = list(
                        name
                        for (
                            name,
                            comp,
                        ) in pyomo.core.base.suffix.active_import_suffix_generator(arg)
                    )
                else:
                    assert isinstance(arg, IBlock)
                    model_suffixes = list(
                        comp.storage_key
                        for comp in pyomo.core.kernel.suffix.import_suffix_generator(
                            arg, active=True, descend_into=False
                        )
                    )

                if len(model_suffixes) > 0:
                    kwds_suffixes = kwds.setdefault('suffixes', [])
                    for name in model_suffixes:
                        if name not in kwds_suffixes:
                            kwds_suffixes.append(name)

        #
        # Handle ephemeral solvers options here. These
        # will override whatever is currently in the options
        # dictionary, but we will reset these options to
        # their original value at the end of this method.
        #

        orig_options = self.options

        self.options = Bunch()
        self.options.update(orig_options)
        self.options.update(kwds.pop('options', {}))
        self.options.update(
            self._options_string_to_dict(kwds.pop('options_string', ''))
        )
        try:
            # we're good to go.
            initial_time = time.time()

            self._presolve(*args, **kwds)

            presolve_completion_time = time.time()
            if self._report_timing:
                print(
                    "      %6.2f seconds required for presolve"
                    % (presolve_completion_time - initial_time)
                )

            if not _model is None:
                self._initialize_callbacks(_model)

            _status = self._apply_solver()
            if hasattr(self, '_transformation_data'):
                del self._transformation_data
            if not hasattr(_status, 'rc'):
                logger.warning(
                    "Solver (%s) did not return a solver status code.\n"
                    "This is indicative of an internal solver plugin error.\n"
                    "Please report this to the Pyomo developers."
                )
            elif _status.rc:
                logger.error(
                    "Solver (%s) returned non-zero return code (%s)"
                    % (self.name, _status.rc)
                )
                if self._tee:
                    logger.error("See the solver log above for diagnostic information.")
                elif hasattr(_status, 'log') and _status.log:
                    logger.error("Solver log:\n" + str(_status.log))
                raise ApplicationError("Solver (%s) did not exit normally" % self.name)
            solve_completion_time = time.time()
            if self._report_timing:
                print(
                    "      %6.2f seconds required for solver"
                    % (solve_completion_time - presolve_completion_time)
                )

            result = self._postsolve()
            result._smap_id = self._smap_id
            result._smap = None
            if _model:
                if isinstance(_model, IBlock):
                    if len(result.solution) == 1:
                        result.solution(0).symbol_map = getattr(
                            _model, "._symbol_maps"
                        )[result._smap_id]
                        result.solution(0).default_variable_value = (
                            self._default_variable_value
                        )
                        if self._load_solutions:
                            _model.load_solution(result.solution(0))
                    else:
                        assert len(result.solution) == 0
                    # see the hack in the write method
                    # we don't want this to stick around on the model
                    # after the solve
                    assert len(getattr(_model, "._symbol_maps")) == 1
                    delattr(_model, "._symbol_maps")
                    del result._smap_id
                    if self._load_solutions and (len(result.solution) == 0):
                        logger.error("No solution is available")
                else:
                    if self._load_solutions:
                        _model.solutions.load_from(
                            result,
                            select=self._select_index,
                            default_variable_value=self._default_variable_value,
                        )
                        result._smap_id = None
                        result.solution.clear()
                    else:
                        result._smap = _model.solutions.symbol_map[self._smap_id]
                        _model.solutions.delete_symbol_map(self._smap_id)
            postsolve_completion_time = time.time()

            if self._report_timing:
                print(
                    "      %6.2f seconds required for postsolve"
                    % (postsolve_completion_time - solve_completion_time)
                )

        finally:
            #
            # Reset the options dict
            #
            self.options = orig_options

        return result

    def _presolve(self, *args, **kwds):
        self._log_file = kwds.pop("logfile", None)
        self._soln_file = kwds.pop("solnfile", None)
        self._select_index = kwds.pop("select", 0)
        self._load_solutions = kwds.pop("load_solutions", True)
        self._timelimit = kwds.pop("timelimit", None)
        self._report_timing = kwds.pop("report_timing", False)
        self._tee = kwds.pop("tee", False)
        self._assert_available = kwds.pop("available", True)
        self._suffixes = kwds.pop("suffixes", [])

        self.available()

        if self._problem_format:
            write_start_time = time.time()
            self._problem_files, self._problem_format, self._smap_id = (
                self._convert_problem(
                    args, self._problem_format, self._valid_problem_formats, **kwds
                )
            )
            total_time = time.time() - write_start_time
            if self._report_timing:
                print("      %6.2f seconds required to write file" % total_time)
        else:
            if len(kwds):
                raise ValueError(
                    "Solver="
                    + self.type
                    + " passed unrecognized keywords: \n\t"
                    + ("\n\t".join("%s = %s" % (k, v) for k, v in kwds.items()))
                )

        if (type(self._problem_files) in (list, tuple)) and (
            not isinstance(self._problem_files[0], str)
        ):
            self._problem_files = self._problem_files[0]._problem_files()
        if self._results_format is None:
            self._results_format = self._default_results_format(self._problem_format)

        #
        # Disabling this check for now.  A solver doesn't have just
        # _one_ results format.
        #
        # if self._results_format not in \
        #   self._valid_result_formats[self._problem_format]:
        #    raise ValueError("Results format '"+str(self._results_format)+"' "
        #                     "cannot be used with problem format '"
        #                     +str(self._problem_format)+"' in solver "+self.name)
        if self._results_format == ResultsFormat.soln:
            self._results_reader = None
        else:
            self._results_reader = pyomo.opt.base.results.ReaderFactory(
                self._results_format
            )

    def _initialize_callbacks(self, model):
        """Initialize call-back functions"""
        pass

    def _apply_solver(self):
        """The routine that performs the solve"""
        raise NotImplementedError  # pragma:nocover

    def _postsolve(self):
        """The routine that does solve post-processing"""
        return self.results

    def _convert_problem(self, args, problem_format, valid_problem_formats, **kwds):
        return convert_problem(
            args, problem_format, valid_problem_formats, self.has_capability, **kwds
        )

    def _default_results_format(self, prob_format):
        """Returns the default results format for different problem
        formats.
        """
        return ResultsFormat.results

    def reset(self):
        """
        Reset the state of the solver
        """
        pass

    def _get_options_string(self, options=None):
        if options is None:
            options = self.options
        ans = []
        for key in options:
            val = options[key]
            if isinstance(val, str) and ' ' in val:
                ans.append("%s=\"%s\"" % (str(key), str(val)))
            else:
                ans.append("%s=%s" % (str(key), str(val)))
        return ' '.join(ans)

    def set_options(self, istr):
        if isinstance(istr, str):
            istr = self._options_string_to_dict(istr)
        for key in istr:
            if not istr[key] is None:
                setattr(self.options, key, istr[key])

    def set_callback(self, name, callback_fn=None):
        """
        Set the callback function for a named callback.

        A call-back function has the form:

            def fn(solver, model):
                pass

        where 'solver' is the native solver interface object and 'model' is
        a Pyomo model instance object.
        """
        if not self._allow_callbacks:
            raise ApplicationError("Callbacks disabled for solver %s" % self.name)
        if callback_fn is None:
            if name in self._callback:
                del self._callback[name]
        else:
            self._callback[name] = callback_fn

    def config_block(self, init=False):
        from pyomo.scripting.solve_config import default_config_block

        return default_config_block(self, init)[0]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________


import logging
import os
import sys
from inspect import currentframe, getframeinfo
from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.log import (
    LegacyPyomoFormatter,
    LoggingIntercept,
    LogHandler,
    LogStream,
    Preformatted,
    StdoutHandler,
    WrappingFormatter,
    is_debug_set,
    pyomo_formatter,
)

logger = logging.getLogger('pyomo.common.log.testing')
filename = getframeinfo(currentframe()).filename


class TestLegacyLogHandler(unittest.TestCase):
    def setUp(self):
        self.stream = StringIO()

    def tearDown(self):
        logger.removeHandler(self.handler)

    def test_simple_log(self):
        # Testing positional base, configurable verbosity
        log = StringIO()
        with LoggingIntercept(log):
            self.handler = LogHandler(
                os.path.dirname(__file__),
                stream=self.stream,
                verbosity=lambda: is_debug_set(logger),
            )
        self.assertIn('LogHandler class has been deprecated', log.getvalue())
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual("", self.stream.getvalue())
        logger.warning("(warn)")
        ans = "WARNING: (warn)\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.warning("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += (
            'WARNING: "[base]%stest_log.py", %d, test_simple_log\n'
            '    (warn)\n' % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_default_verbosity(self):
        # Testing positional base, configurable verbosity
        log = StringIO()
        with LoggingIntercept(log):
            self.handler = LogHandler(os.path.dirname(__file__), stream=self.stream)
        self.assertIn('LogHandler class has been deprecated', log.getvalue())
        logger.addHandler(self.handler)

        logger.setLevel(logging.WARNING)
        logger.warning("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans = (
            'WARNING: "[base]%stest_log.py", %d, test_default_verbosity\n'
            '    (warn)\n' % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())


class TestWrappingFormatter(unittest.TestCase):
    def setUp(self):
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        logger.addHandler(self.handler)

    def tearDown(self):
        logger.removeHandler(self.handler)

    def test_style_options(self):
        ans = ''

        self.handler.setFormatter(WrappingFormatter(style='%'))
        logger.warning("(warn)")
        ans += "WARNING: (warn)\n"
        self.assertEqual(ans, self.stream.getvalue())

        self.handler.setFormatter(WrappingFormatter(style='$'))
        logger.warning("(warn)")
        ans += "WARNING: (warn)\n"
        self.assertEqual(ans, self.stream.getvalue())

        self.handler.setFormatter(WrappingFormatter(style='{'))
        logger.warning("(warn)")
        ans += "WARNING: (warn)\n"
        self.assertEqual(ans, self.stream.getvalue())

        with self.assertRaisesRegex(ValueError, 'unrecognized style flag "s"'):
            WrappingFormatter(style='s')


class TestLegacyPyomoFormatter(unittest.TestCase):
    def setUp(self):
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        logger.addHandler(self.handler)

    def tearDown(self):
        logger.removeHandler(self.handler)

    def test_unallowed_options(self):
        with self.assertRaisesRegex(ValueError, "'fmt' is not a valid option"):
            LegacyPyomoFormatter(fmt='%(message)')

        with self.assertRaisesRegex(ValueError, "'style' is not a valid option"):
            LegacyPyomoFormatter(style='%')

    def test_simple_log(self):
        # Testing positional base, configurable verbosity
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual("", self.stream.getvalue())
        logger.warning("(warn)")
        ans = "WARNING: (warn)\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.warning("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += (
            'WARNING: "[base]%stest_log.py", %d, test_simple_log\n'
            '    (warn)\n' % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_alternate_base(self):
        self.handler.setFormatter(LegacyPyomoFormatter(base='log_config'))

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual("", self.stream.getvalue())
        logger.warning("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans = 'WARNING: "%s", %d, test_alternate_base\n    (warn)\n' % (
            filename,
            lineno,
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_no_base(self):
        self.handler.setFormatter(LegacyPyomoFormatter())

        logger.setLevel(logging.WARNING)
        logger.info("(info)")
        self.assertEqual("", self.stream.getvalue())
        logger.warning("(warn)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans = 'WARNING: "%s", %d, test_no_base\n    (warn)\n' % (filename, lineno)
        self.assertEqual(ans, self.stream.getvalue())

    def test_no_message(self):
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        logger.setLevel(logging.WARNING)
        logger.info("")
        self.assertEqual("", self.stream.getvalue())

        logger.warning("")
        ans = "WARNING:\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.warning("")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'WARNING: "[base]%stest_log.py", %d, test_no_message\n\n' % (
            os.path.sep,
            lineno,
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_blank_lines(self):
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        logger.setLevel(logging.WARNING)
        logger.warning("\n\nthis is a message.\n\n\n")
        ans = "WARNING: this is a message.\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.warning("\n\nthis is a message.\n\n\n")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += (
            'WARNING: "[base]%stest_log.py", %d, test_blank_lines\n'
            "    this is a message.\n" % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_numbered_level(self):
        testname = 'test_numbered_level'
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        logger.setLevel(logging.WARNING)
        logger.log(45, "(hi)")
        ans = "Level 45: (hi)\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.log(45, "")
        ans += "Level 45:\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.log(45, "(hi)")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'Level 45: "[base]%stest_log.py", %d, %s\n    (hi)\n' % (
            os.path.sep,
            lineno,
            testname,
        )
        self.assertEqual(ans, self.stream.getvalue())

        logger.log(45, "")
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'Level 45: "[base]%stest_log.py", %d, %s\n\n' % (
            os.path.sep,
            lineno,
            testname,
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_preformatted(self):
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        msg = """This is a long multi-line message that in normal circumstances \
would be line-wrapped
        with additional information
        that normally would be combined."""

        logger.setLevel(logging.WARNING)
        logger.info(msg)
        self.assertEqual("", self.stream.getvalue())

        logger.warning(Preformatted(msg))
        ans = msg + "\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.warning(msg)
        ans += (
            "WARNING: This is a long multi-line message that in normal "
            "circumstances would\n"
            "be line-wrapped with additional information that normally would be combined.\n"
        )
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)

        logger.warning(Preformatted(msg))
        ans += msg + "\n"
        self.assertEqual(ans, self.stream.getvalue())

        logger.warning(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += 'WARNING: "[base]%stest_log.py", %d, test_preformatted\n' % (
            os.path.sep,
            lineno,
        )
        ans += (
            "    This is a long multi-line message that in normal "
            "circumstances would be\n"
            "    line-wrapped with additional information that normally would be combined.\n"
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_long_messages(self):
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        msg = (
            "This is a long message\n"
            "\n"
            "With some kind of internal formatting\n"
            "    - including a bulleted list\n"
            "    - list 2  "
        )
        logger.setLevel(logging.WARNING)
        logger.warning(msg)
        ans = (
            "WARNING: This is a long message\n"
            "\n"
            "    With some kind of internal formatting\n"
            "        - including a bulleted list\n"
            "        - list 2\n"
        )
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.info(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += (
            'INFO: "[base]%stest_log.py", %d, test_long_messages\n'
            "    This is a long message\n"
            "\n"
            "    With some kind of internal formatting\n"
            "        - including a bulleted list\n"
            "        - list 2\n" % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())

        # test trailing newline
        msg += "\n"
        logger.setLevel(logging.WARNING)
        logger.warning(msg)
        ans += (
            "WARNING: This is a long message\n"
            "\n"
            "    With some kind of internal formatting\n"
            "        - including a bulleted list\n"
            "        - list 2\n"
        )
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.info(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += (
            'INFO: "[base]%stest_log.py", %d, test_long_messages\n'
            "    This is a long message\n"
            "\n"
            "    With some kind of internal formatting\n"
            "        - including a bulleted list\n"
            "        - list 2\n" % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())

        # test initial and final blank lines
        msg = "\n" + msg + "\n\n"
        logger.setLevel(logging.WARNING)
        logger.warning(msg)
        ans += (
            "WARNING: This is a long message\n"
            "\n"
            "    With some kind of internal formatting\n"
            "        - including a bulleted list\n"
            "        - list 2\n"
        )
        self.assertEqual(ans, self.stream.getvalue())

        logger.setLevel(logging.DEBUG)
        logger.info(msg)
        lineno = getframeinfo(currentframe()).lineno - 1
        ans += (
            'INFO: "[base]%stest_log.py", %d, test_long_messages\n'
            "    This is a long message\n"
            "\n"
            "    With some kind of internal formatting\n"
            "        - including a bulleted list\n"
            "        - list 2\n" % (os.path.sep, lineno)
        )
        self.assertEqual(ans, self.stream.getvalue())

    def test_verbatim(self):
        self.handler.setFormatter(
            LegacyPyomoFormatter(
                base=os.path.dirname(__file__), verbosity=lambda: is_debug_set(logger)
            )
        )

        msg = (
            "This is a long message\n"
            "\n"
            "   ```\n"
            "With some \n"
            "internal\n"
            "verbatim \n"
            "  - including a\n"
            "    long list\n"
            "  - and a short list \n"
            "  ```\n"
            "\n"
            "And some \n"
            "internal\n"
            "non-verbatim \n"
            "  - including a\n"
            "    long list\n"
            "  - and a short list \n"
            "\n"
            "And a section\n"
            "~~~~~~~~~~~~~\n"
            "\n"
            "  | and\n"
            "  | a line\n"
            "  | block\n"
            "\n"
            "And a\n"
            "quoted literal::\n"
            "\n"
            ">> he said\n"
            ">\n"
            "> and they replied\n"
            "\n"
            "this is\n"
            "outside the quote\n"
            "\n"
            "indented literal::\n"
            "\n"
            "    Here is\n"
            "       an indented\n"
            "\n"
            "    literal\n"
            "    with a blank line\n"
            "\n"
            "Finally, an invalid::\n"
            "\n"
            "quote\n"
            "block\n"
        )
        logger.setLevel(logging.WARNING)
        logger.warning(msg)
        ans = (
            "WARNING: This is a long message\n"
            "\n"
            "    With some \n"
            "    internal\n"
            "    verbatim \n"
            "      - including a\n"
            "        long list\n"
            "      - and a short list \n"
            "\n"
            "    And some internal non-verbatim\n"
            "      - including a long list\n"
            "      - and a short list\n"
            "\n"
            "    And a section\n"
            "    ~~~~~~~~~~~~~\n"
            "\n"
            "      | and\n"
            "      | a line\n"
            "      | block\n"
            "\n"
            "    And a quoted literal::\n"
            "\n"
            "    >> he said\n"
            "    >\n"
            "    > and they replied\n"
            "\n"
            "    this is outside the quote\n"
            "\n"
            "    indented literal::\n"
            "\n"
            "        Here is\n"
            "           an indented\n"
            "\n"
            "        literal\n"
            "        with a blank line\n"
            "\n"
            "    Finally, an invalid::\n"
            "\n"
            "    quote block\n"
        )
        self.assertEqual(ans, self.stream.getvalue())


class TestLogStream(unittest.TestCase):
    def test_log_stream(self):
        ls = LogStream(logging.INFO, logging.getLogger('pyomo'))
        LI = LoggingIntercept(level=logging.INFO, formatter=pyomo_formatter)
        with LI as OUT:
            ls.write("hello, world\n")
            self.assertEqual(OUT.getvalue(), "INFO: hello, world\n")

        with LI as OUT:
            ls.write("line 1\nline 2\n")
            self.assertEqual(OUT.getvalue(), "INFO: line 1\nINFO: line 2\n")

        with LI as OUT:
            # empty writes do not generate log records
            ls.write("")
            ls.flush()
            self.assertEqual("", OUT.getvalue())

        with LI as OUT:
            ls.write("line 1\nline 2")
            self.assertEqual(OUT.getvalue(), "INFO: line 1\n")

        with LI as OUT:
            ls.flush()
            self.assertEqual(OUT.getvalue(), "INFO: line 2\n")
            # Second flush should do nothing
            ls.flush()
            self.assertEqual(OUT.getvalue(), "INFO: line 2\n")

        with LI as OUT:
            with LogStream(logging.INFO, logging.getLogger('pyomo')) as ls:
                ls.write('line 1\nline 2')
                self.assertEqual(OUT.getvalue(), "INFO: line 1\n")
            # Exiting the context manager flushes the LogStream
            self.assertEqual(OUT.getvalue(), "INFO: line 1\nINFO: line 2\n")

    def test_loggerAdapter(self):
        class Adapter(logging.LoggerAdapter):
            def process(self, msg, kwargs):
                return '[%s] %s' % (self.extra['foo'], msg), kwargs

        adapter = Adapter(logging.getLogger('pyomo'), {"foo": 42})
        ls = LogStream(logging.INFO, adapter)
        LI = LoggingIntercept(level=logging.INFO, formatter=pyomo_formatter)
        with LI as OUT:
            ls.write("hello, world\n")
            self.assertEqual(OUT.getvalue(), "INFO: [42] hello, world\n")


class TestPreformatted(unittest.TestCase):
    def test_preformatted_api(self):
        ref = 'a message'
        msg = Preformatted(ref)
        self.assertIs(msg.msg, ref)
        self.assertEqual(str(msg), ref)
        self.assertEqual(repr(msg), "Preformatted('a message')")

        ref = 2
        msg = Preformatted(ref)
        self.assertIs(msg.msg, ref)
        self.assertEqual(str(msg), '2')
        self.assertEqual(repr(msg), "Preformatted(2)")


class TestLoggingIntercept(unittest.TestCase):
    def test_init(self):
        li = LoggingIntercept()
        self.assertEqual(li.module, 'root')

        li = LoggingIntercept(module='pyomo.core')
        self.assertEqual(li.module, 'pyomo.core')

        li = LoggingIntercept(logger=logger)
        self.assertEqual(li.module, 'pyomo.common.log.testing')

        with self.assertRaisesRegex(
            ValueError, "LoggingIntercept: only one of 'module' and 'logger' is allowed"
        ):
            li = LoggingIntercept(module='pyomo', logger=logger)

    def test_propagate(self):
        self.assertEqual(logger.propagate, True)
        with LoggingIntercept(logger=logger):
            self.assertEqual(logger.propagate, False)
        self.assertEqual(logger.propagate, True)

    def test_propagate(self):
        self.assertEqual(logger.level, 30)
        try:
            with LoggingIntercept(logger=logger, level=None):
                self.assertEqual(logger.level, 30)
            self.assertEqual(logger.level, 30)
            with LoggingIntercept(logger=logger, level=40):
                self.assertEqual(logger.level, 40)
            self.assertEqual(logger.level, 30)

            logger.setLevel(40)
            with LoggingIntercept(logger=logger, level=None):
                self.assertEqual(logger.level, 40)
            self.assertEqual(logger.level, 40)
            with LoggingIntercept(logger=logger, level=30):
                self.assertEqual(logger.level, 30)
            self.assertEqual(logger.level, 40)
        finally:
            logger.setLevel(30)


class TestStdoutHandler(unittest.TestCase):
    def setUp(self):
        self.orig = sys.stdout

    def tearDown(self):
        sys.stdout = self.orig

    def test_emit(self):
        handler = StdoutHandler()
        self.assertIsNone(handler.stream)

        sys.stdout = StringIO()
        record = logging.LogRecord(
            name="test",
            level=logging.WARNING,
            pathname="/path/to/file.py",
            lineno=42,
            msg="Test msg",
            args=(),
            exc_info=None,
        )
        handler.emit(record)
        handler.flush()

        self.assertEqual(sys.stdout.getvalue(), "Test msg\n")
        self.assertIsNone(handler.stream)

    def test_handler(self):
        logger = logging.getLogger(__name__)
        propagate, level = logger.propagate, logger.level
        handler = StdoutHandler()
        try:
            logger.addHandler(handler)
            logger.propagate = False
            sys.stdout = StringIO()

            logger.setLevel(logging.WARNING)
            logger.info("Test1")
            self.assertEqual("", sys.stdout.getvalue())

            logger.setLevel(logging.INFO)
            logger.info("Test2")
            self.assertEqual(sys.stdout.getvalue(), "Test2\n")
        finally:
            logger.removeHandler(handler)
            logger.propagate = propagate
            logger.setLevel(level)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Functions for construction and solution of the PyROS master problem.
"""

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.modeling import unique_component_name
from pyomo.core import TransformationFactory
from pyomo.core.base import ConcreteModel, Block, Var, Objective, Constraint
from pyomo.core.base.set_types import NonNegativeIntegers, NonNegativeReals
from pyomo.core.expr import identify_variables, value
from pyomo.core.util import prod
from pyomo.opt import TerminationCondition as tc
from pyomo.repn.standard_repn import generate_standard_repn

from pyomo.contrib.pyros.solve_data import MasterResults
from pyomo.contrib.pyros.util import (
    call_solver,
    DR_POLISHING_PARAM_PRODUCT_ZERO_TOL,
    enforce_dr_degree,
    get_all_first_stage_eq_cons,
    get_dr_expression,
    check_time_limit_reached,
    generate_all_decision_rule_var_data_objects,
    ObjectiveType,
    pyrosTerminationCondition,
    TIC_TOC_SOLVE_TIME_ATTR,
    write_subproblem,
)


def construct_initial_master_problem(model_data):
    """
    Construct the initial master problem model object
    from the preprocessed working model.

    Parameters
    ----------
    model_data : model data object
        Main model data object,
        containing the preprocessed working model.

    Returns
    -------
    master_model : ConcreteModel
        Initial master problem model object.
        Contains a single scenario block fully cloned from
        the working model.
    """
    master_model = ConcreteModel()
    master_model.scenarios = Block(NonNegativeIntegers, NonNegativeIntegers)
    add_scenario_block_to_master_problem(
        master_model=master_model,
        scenario_idx=(0, 0),
        param_realization=model_data.config.nominal_uncertain_param_vals,
        from_block=model_data.working_model,
        clone_first_stage_components=True,
    )

    # epigraph Objective was not added during preprocessing,
    # as we wanted to add it to the root block of the master
    # model rather than to the model to prevent
    # duplication across scenario sub-blocks
    master_model.epigraph_obj = Objective(
        expr=master_model.scenarios[0, 0].first_stage.epigraph_var
    )

    return master_model


def add_scenario_block_to_master_problem(
    master_model,
    scenario_idx,
    param_realization,
    from_block,
    clone_first_stage_components,
):
    """
    Add new scenario block to the master model.

    Parameters
    ----------
    master_model : ConcreteModel
        Master model.
    scenario_idx : tuple
        Index of ``master_model.scenarios`` for the new block.
    param_realization : Iterable of numeric type
        Uncertain parameter realization for new block.
    from_block : BlockData
        Block from which to transfer attributes.
        This can be an existing scenario block, or a block
        with the same hierarchical structure as the
        preprocessed working model.
    clone_first_stage_components : bool
        True to clone first-stage variables
        when transferring attributes to the new block
        to the new block (as opposed to using the objects as
        they are in `from_block`), False otherwise.
    """
    # Note for any of the Vars not copied:
    # - if Var is not a member of an indexed var, then
    #   the 'name' attribute changes from
    #   '{from_block.name}.{var.name}'
    #   to 'scenarios[{scenario_idx}].{var.name}'
    # - otherwise, the name stays the same
    memo = dict()
    if not clone_first_stage_components:
        nonadjustable_comps = from_block.all_nonadjustable_variables
        memo = {id(comp): comp for comp in nonadjustable_comps}

        # we will clone the first-stage constraints
        # (mostly to prevent symbol map name clashes).
        # the duplicate constraints are redundant.
        # consider deactivating these constraints in the
        # off-nominal blocks?

    new_block = from_block.clone(memo=memo)
    master_model.scenarios[scenario_idx].transfer_attributes_from(new_block)

    # update uncertain parameter values in new block
    new_uncertain_params = master_model.scenarios[scenario_idx].uncertain_params
    for param, val in zip(new_uncertain_params, param_realization):
        param.set_value(val)

    # deactivate the first-stage constraints: they are duplicate
    if scenario_idx != (0, 0):
        new_blk = master_model.scenarios[scenario_idx]
        for con in new_blk.first_stage.inequality_cons.values():
            con.deactivate()
        for con in get_all_first_stage_eq_cons(new_blk):
            con.deactivate()


def construct_master_feasibility_problem(master_data):
    """
    Construct slack variable minimization problem from the master
    model.

    Slack variables are added only to the seconds-stage
    inequality constraints of the blocks added for the
    current PyROS iteration.

    Parameters
    ----------
    master_data : MasterProblemData
        Master problem data.

    Returns
    -------
    slack_model : ConcreteModel
        Slack variable model.
    """
    # to prevent use of find_component when copying variable values
    # from the slack model to the master problem later, we will
    # map corresponding variables before/during slack model construction
    varmap_name = unique_component_name(master_data.master_model, 'pyros_var_map')
    setattr(
        master_data.master_model,
        varmap_name,
        list(master_data.master_model.component_data_objects(Var)),
    )

    slack_model = master_data.master_model.clone()

    master_data.feasibility_problem_varmap = list(
        zip(
            getattr(master_data.master_model, varmap_name),
            getattr(slack_model, varmap_name),
        )
    )
    delattr(master_data.master_model, varmap_name)
    delattr(slack_model, varmap_name)

    for obj in slack_model.component_data_objects(Objective):
        obj.deactivate()
    iteration = master_data.iteration

    # add slacks only to second-stage inequality constraints for the
    # newest master block(s).
    # these should be the only constraints that
    # may have been violated by the previous master and separation
    # solution(s)
    targets = []
    for blk in slack_model.scenarios[iteration, :]:
        targets.extend(blk.second_stage.inequality_cons.values())

    # retain original constraint expressions before adding slacks
    # (to facilitate slack initialization and scaling)
    pre_slack_con_exprs = ComponentMap((con, con.body - con.upper) for con in targets)

    # add slack variables and objective
    # inequalities g(v) <= b become g(v) - s^- <= b
    TransformationFactory("core.add_slack_variables").apply_to(
        slack_model, targets=targets
    )
    slack_vars = ComponentSet(
        slack_model._core_add_slack_variables.component_data_objects(
            Var, descend_into=True
        )
    )

    # initialize slack variables
    for con in pre_slack_con_exprs:
        # get mapping from slack variables to their (linear)
        # coefficients (+/-1) in the updated constraint expressions
        repn = generate_standard_repn(con.body)
        slack_var_coef_map = ComponentMap()
        for idx in range(len(repn.linear_vars)):
            var = repn.linear_vars[idx]
            if var in slack_vars:
                slack_var_coef_map[var] = repn.linear_coefs[idx]

        for slack_var in slack_var_coef_map:
            # coefficient determines whether the slack
            # is a +ve or -ve slack
            if slack_var_coef_map[slack_var] == -1:
                con_slack = max(0, value(pre_slack_con_exprs[con]))
            else:
                con_slack = max(0, -value(pre_slack_con_exprs[con]))

            slack_var.set_value(con_slack)

    return slack_model


def solve_master_feasibility_problem(master_data):
    """
    Solve a slack variable-based feasibility model derived
    from the master problem. Initialize the master problem
    to the  solution found by the optimizer if solved successfully,
    or to the initial point provided to the solver otherwise.

    Parameters
    ----------
    master_data : MasterProblemData
        Master problem data.

    Returns
    -------
    results : SolverResults
        Solver results.
    """
    model = construct_master_feasibility_problem(master_data)

    active_obj = next(model.component_data_objects(Objective, active=True))

    config = master_data.config
    config.progress_logger.debug("Solving master feasibility problem")
    config.progress_logger.debug(
        f" Initial objective (total slack): {value(active_obj)}"
    )

    if config.solve_master_globally:
        solver = config.global_solver
    else:
        solver = config.local_solver

    results = call_solver(
        model=model,
        solver=solver,
        config=config,
        timing_obj=master_data.timing,
        timer_name="main.master_feasibility",
        err_msg=(
            f"Optimizer {repr(solver)} encountered exception "
            "attempting to solve master feasibility problem in iteration "
            f"{master_data.iteration}."
        ),
    )

    feasible_terminations = {
        tc.optimal,
        tc.locallyOptimal,
        tc.globallyOptimal,
        tc.feasible,
    }
    if results.solver.termination_condition in feasible_terminations:
        model.solutions.load_from(results)
        config.progress_logger.debug(
            f" Final objective (total slack): {value(active_obj)}"
        )
        config.progress_logger.debug(
            f" Termination condition: {results.solver.termination_condition}"
        )
        config.progress_logger.debug(
            f" Solve time: {getattr(results.solver, TIC_TOC_SOLVE_TIME_ATTR)}s"
        )
    else:
        config.progress_logger.debug(
            "Could not successfully solve master feasibility problem "
            f"of iteration {master_data.iteration} with primary subordinate "
            f"{'global' if config.solve_master_globally else 'local'} solver "
            "to acceptable level. "
            f"Termination stats:\n{results.solver}\n"
            "Maintaining unoptimized point for master problem initialization."
        )

    # load master feasibility point to master model
    for master_var, feas_var in master_data.feasibility_problem_varmap:
        master_var.set_value(feas_var.value, skip_validation=True)

    return results


def construct_dr_polishing_problem(master_data):
    """
    Construct DR polishing problem from the master problem.

    Parameters
    ----------
    master_data : MasterProblemData
        Master problem data.

    Returns
    -------
    polishing_model : ConcreteModel
        Polishing model.

    Note
    ----
    Polishing problem is to minimize the L1-norm of the vector of
    all decision rule polynomial terms, subject to the original
    master problem constraints, with all first-stage variables
    (including epigraph) fixed. Optimality of the polished
    DR with respect to the master objective is also enforced.
    """
    master_model = master_data.master_model
    polishing_model = master_model.clone()
    nominal_polishing_block = polishing_model.scenarios[0, 0]

    nominal_eff_var_partitioning = nominal_polishing_block.effective_var_partitioning

    nondr_nonadjustable_vars = (
        nominal_eff_var_partitioning.first_stage_variables
        # fixing epigraph variable constrains the problem
        # to the optimal master problem solution set
        + [nominal_polishing_block.first_stage.epigraph_var]
    )
    for var in nondr_nonadjustable_vars:
        var.fix()

    # deactivate original constraints that involved
    # only vars that have been fixed.
    # we do this mostly to ensure that the active equality constraints
    # do not grossly outnumber the unfixed Vars
    fixed_dr_vars = [
        var
        for var in generate_all_decision_rule_var_data_objects(nominal_polishing_block)
        if var.fixed
    ]
    fixed_nonadjustable_vars = ComponentSet(nondr_nonadjustable_vars + fixed_dr_vars)
    for blk in polishing_model.scenarios.values():
        for con in blk.component_data_objects(Constraint, active=True):
            vars_in_con = ComponentSet(identify_variables(con.body))
            if not (vars_in_con - fixed_nonadjustable_vars):
                con.deactivate()

    # we will add the polishing objective later
    polishing_model.epigraph_obj.deactivate()

    polishing_model.polishing_vars = polishing_vars = []
    indexed_dr_var_list = nominal_polishing_block.first_stage.decision_rule_vars
    for idx, indexed_dr_var in enumerate(indexed_dr_var_list):
        # auxiliary 'polishing' variables.
        # these are meant to represent the absolute values
        # of the terms of DR polynomial;
        # we need these for the L1-norm
        indexed_polishing_var = Var(
            list(indexed_dr_var.keys()), domain=NonNegativeReals
        )
        polishing_model.add_component(f"dr_polishing_var_{idx}", indexed_polishing_var)
        polishing_vars.append(indexed_polishing_var)

    # we need the DR expressions to set up the
    # absolute value constraints and initialize the
    # auxiliary polishing variables
    eff_ss_var_to_dr_expr_pairs = [
        (ss_var, get_dr_expression(nominal_polishing_block, ss_var))
        for ss_var in nominal_eff_var_partitioning.second_stage_variables
    ]

    dr_eq_var_zip = zip(polishing_vars, eff_ss_var_to_dr_expr_pairs)
    polishing_model.polishing_abs_val_lb_cons = all_lb_cons = []
    polishing_model.polishing_abs_val_ub_cons = all_ub_cons = []
    for idx, (indexed_polishing_var, (ss_var, dr_expr)) in enumerate(dr_eq_var_zip):
        # set up absolute value constraint components
        polishing_absolute_value_lb_cons = Constraint(indexed_polishing_var.index_set())
        polishing_absolute_value_ub_cons = Constraint(indexed_polishing_var.index_set())

        # add indexed constraints to polishing model
        polishing_model.add_component(
            f"polishing_abs_val_lb_con_{idx}", polishing_absolute_value_lb_cons
        )
        polishing_model.add_component(
            f"polishing_abs_val_ub_con_{idx}", polishing_absolute_value_ub_cons
        )

        # update list of absolute value (i.e., polishing) cons
        all_lb_cons.append(polishing_absolute_value_lb_cons)
        all_ub_cons.append(polishing_absolute_value_ub_cons)

        for dr_monomial in dr_expr.args:
            is_a_nonstatic_dr_term = dr_monomial.is_expression_type()
            if is_a_nonstatic_dr_term:
                # degree >= 1 monomial expression of form
                # (product of uncertain params) * dr variable
                dr_var_in_term = dr_monomial.args[-1]
            else:
                # the static term (intercept)
                dr_var_in_term = dr_monomial

            # we want the DR variable and corresponding polishing
            # constraints to have the same index in the indexed
            # components
            dr_var_in_term_idx = dr_var_in_term.index()
            polishing_var = indexed_polishing_var[dr_var_in_term_idx]

            # Fix DR variable if:
            # (1) it has already been fixed from master due to
            #     DR efficiencies (already done)
            # (2) coefficient of term
            #     (i.e. product of uncertain parameter values)
            #     in DR expression is 0
            #     across all master blocks
            dr_term_copies = [
                (
                    scenario_blk.second_stage.decision_rule_eqns[idx].body.args[
                        dr_var_in_term_idx
                    ]
                )
                for scenario_blk in master_model.scenarios.values()
            ]
            all_copy_coeffs_zero = is_a_nonstatic_dr_term and all(
                abs(value(prod(term.args[:-1]))) <= DR_POLISHING_PARAM_PRODUCT_ZERO_TOL
                for term in dr_term_copies
            )
            if all_copy_coeffs_zero:
                # increment static DR variable value
                # to maintain feasibility of the initial point
                # as much as possible
                static_dr_var_in_expr = dr_expr.args[0]
                static_dr_var_in_expr.set_value(
                    value(static_dr_var_in_expr) + value(dr_monomial)
                )
                dr_var_in_term.fix(0)

            # add polishing constraints
            polishing_absolute_value_lb_cons[dr_var_in_term_idx] = (
                -polishing_var - dr_monomial <= 0
            )
            polishing_absolute_value_ub_cons[dr_var_in_term_idx] = (
                dr_monomial - polishing_var <= 0
            )

            # some DR variables may be fixed,
            # due to the PyROS DR order efficiency instituted
            # in the first few iterations.
            # these need not be polished
            if dr_var_in_term.fixed or not is_a_nonstatic_dr_term:
                polishing_var.fix()
                polishing_absolute_value_lb_cons[dr_var_in_term_idx].deactivate()
                polishing_absolute_value_ub_cons[dr_var_in_term_idx].deactivate()

            # ensure polishing var properly initialized
            polishing_var.set_value(abs(value(dr_monomial)))

    # L1-norm objective
    # TODO: if dropping nonstatic terms, ensure the
    #       corresponding polishing variables are excluded
    #       from this expression
    polishing_model.polishing_obj = Objective(
        expr=sum(sum(polishing_var.values()) for polishing_var in polishing_vars)
    )

    return polishing_model


def minimize_dr_vars(master_data):
    """
    Polish decision rule of most recent master problem solution.

    Parameters
    ----------
    master_data : MasterProblemData
        Master problem data.

    Returns
    -------
    results : SolverResults
        Subordinate solver results for the polishing problem.
    polishing_successful : bool
        True if polishing model was solved to acceptable level,
        False otherwise.
    """
    config = master_data.config

    # create polishing NLP
    polishing_model = construct_dr_polishing_problem(master_data)

    if config.solve_master_globally:
        solver = config.global_solver
    else:
        solver = config.local_solver

    config.progress_logger.debug("Solving DR polishing problem")

    # polishing objective should be consistent with value of sum
    # of absolute values of polynomial DR terms provided
    # auxiliary variables initialized correctly
    polishing_obj = polishing_model.polishing_obj
    config.progress_logger.debug(f" Initial DR norm: {value(polishing_obj)}")

    # === Solve the polishing model
    results = call_solver(
        model=polishing_model,
        solver=solver,
        config=config,
        timing_obj=master_data.timing,
        timer_name="main.dr_polishing",
        err_msg=(
            f"Optimizer {repr(solver)} encountered an exception "
            "attempting to solve decision rule polishing problem "
            f"in iteration {master_data.iteration}"
        ),
    )

    # interested in the time and termination status for debugging
    # purposes
    config.progress_logger.debug(" Done solving DR polishing problem")

    # === Process solution by termination condition
    acceptable = {tc.globallyOptimal, tc.optimal, tc.locallyOptimal}
    if results.solver.termination_condition not in acceptable:
        # continue with "unpolished" master model solution
        config.progress_logger.debug(
            "Could not successfully solve DR polishing problem "
            f"of iteration {master_data.iteration} with primary subordinate "
            f"{'global' if config.solve_master_globally else 'local'} solver "
            "to acceptable level. "
            f"Termination stats:\n{results.solver}\n"
            "Maintaining unpolished master problem solution."
        )
        return results, False

    # update master model second-stage, state, and decision rule
    # variables to polishing model solution
    polishing_model.solutions.load_from(results)

    # update master problem variable values
    for idx, blk in master_data.master_model.scenarios.items():
        master_adjustable_vars = blk.all_adjustable_variables
        polishing_adjustable_vars = polishing_model.scenarios[
            idx
        ].all_adjustable_variables
        adjustable_vars_zip = zip(master_adjustable_vars, polishing_adjustable_vars)
        for master_var, polish_var in adjustable_vars_zip:
            master_var.set_value(value(polish_var))
        dr_var_zip = zip(
            blk.first_stage.decision_rule_vars,
            polishing_model.scenarios[idx].first_stage.decision_rule_vars,
        )
        for master_dr, polish_dr in dr_var_zip:
            for mvar, pvar in zip(master_dr.values(), polish_dr.values()):
                mvar.set_value(value(pvar), skip_validation=True)

    config.progress_logger.debug(f" Optimized DR norm: {value(polishing_obj)}")
    log_master_solve_results(polishing_model, config, results, desc="polished")

    return results, True


def get_master_dr_degree(master_data):
    """
    Determine DR polynomial degree to enforce based on
    the iteration number and/or the presence of first-stage
    equality constraints that depend on the decision rule variables.

    If there are first-stage equality constraints that depend
    on the decision rule variables, such as equalities derived
    from coefficient matching or discretization of state-variable
    independent equalities, then the degree is set to
    ``config.decision_rule_order``.

    Otherwise, the degree is set to:

    - 0 if iteration number is 0
    - min(1, config.decision_rule_order) if iteration number
      otherwise does not exceed number of effective
      uncertain parameters
    - min(2, config.decision_rule_order) otherwise.

    Parameters
    ----------
    master_data : MasterProblemData
        Master problem data.

    Returns
    -------
    int
        DR order, or polynomial degree, to enforce.
    """
    nom_scenario_blk = master_data.master_model.scenarios[0, 0]
    if nom_scenario_blk.first_stage.dr_dependent_equality_cons:
        return master_data.config.decision_rule_order

    if master_data.iteration == 0:
        return 0
    elif master_data.iteration <= len(nom_scenario_blk.effective_uncertain_params):
        return min(1, master_data.config.decision_rule_order)
    else:
        return min(2, master_data.config.decision_rule_order)


def higher_order_decision_rule_efficiency(master_data):
    """
    Enforce DR coefficient variable efficiencies for
    master problem-like formulation.

    Parameters
    ----------
    master_data : MasterProblemData
        Master problem data.

    Note
    ----
    The DR coefficient variable efficiencies consist of
    setting the degree of the DR polynomial expressions
    by fixing the appropriate variables to 0. The degree
    to be set depends on the iteration number;
    see ``get_master_dr_degree``.
    """
    order_to_enforce = get_master_dr_degree(master_data)
    enforce_dr_degree(
        working_blk=master_data.master_model.scenarios[0, 0],
        config=master_data.config,
        degree=order_to_enforce,
    )


def log_master_solve_results(master_model, config, results, desc="Optimized"):
    """
    Log master problem solve results.
    """
    if config.objective_focus == ObjectiveType.worst_case:
        eval_obj_blk_idx = max(
            master_model.scenarios.keys(),
            key=lambda idx: value(master_model.scenarios[idx].second_stage_objective),
        )
    else:
        eval_obj_blk_idx = (0, 0)

    eval_obj_blk = master_model.scenarios[eval_obj_blk_idx]
    config.progress_logger.debug(f" {desc.capitalize()} master objective breakdown:")
    config.progress_logger.debug(
        f"  First-stage objective: {value(eval_obj_blk.first_stage_objective)}"
    )
    config.progress_logger.debug(
        f"  Second-stage objective: {value(eval_obj_blk.second_stage_objective)}"
    )
    master_obj = eval_obj_blk.full_objective
    config.progress_logger.debug(f"  Overall Objective: {value(master_obj)}")
    config.progress_logger.debug(
        f" Termination condition: {results.solver.termination_condition}"
    )
    config.progress_logger.debug(
        f" Solve time: {getattr(results.solver, TIC_TOC_SOLVE_TIME_ATTR)}s"
    )


def process_termination_condition_master_problem(config, results):
    """
    Process master problem solve termination condition.

    Parameters
    ----------
    config : ConfigDict
        PyROS solver options.
    results : SolverResults
        Solver results.

    Returns
    -------
    optimality_acceptable : bool
        True if problem was solved to an acceptable optimality target,
        False otherwise.
    infeasible : bool
        True if problem was found to be infeasible, False otherwise.

    Raises
    ------
    NotImplementedError
        If a particular solver termination is not supported by
        PyROS.
    """
    locally_acceptable = [tc.optimal, tc.locallyOptimal, tc.globallyOptimal]
    globally_acceptable = [tc.optimal, tc.globallyOptimal]
    robust_infeasible = [tc.infeasible]
    try_backups = [
        tc.feasible,
        tc.maxTimeLimit,
        tc.maxIterations,
        tc.maxEvaluations,
        tc.minStepLength,
        tc.minFunctionValue,
        tc.other,
        tc.solverFailure,
        tc.internalSolverError,
        tc.error,
        tc.unbounded,
        tc.infeasibleOrUnbounded,
        tc.invalidProblem,
        tc.intermediateNonInteger,
        tc.noSolution,
        tc.unknown,
    ]

    termination_condition = results.solver.termination_condition
    optimality_acceptable = (
        (termination_condition in globally_acceptable)
        if config.solve_master_globally
        else (termination_condition in locally_acceptable)
    )
    infeasible = termination_condition in robust_infeasible
    try_backup_solver = termination_condition in try_backups

    unsupported_termination = not (
        optimality_acceptable or try_backup_solver or infeasible
    )
    if unsupported_termination:
        solve_type = "global" if config.solve_master_globally else "local"
        raise NotImplementedError(
            f"Processing of termination condition {termination_condition} "
            f"for attempt at {solve_type} solution of master problem "
            "is currently not supported by PyROS. "
            "Please report this issue to the PyROS developers."
        )

    return optimality_acceptable, infeasible


def solver_call_master(master_data):
    """
    Invoke subsolver(s) on PyROS master problem,
    and update the MasterResults object accordingly.

    Parameters
    ----------
    master_data : MasterProblemData
        Container for current master problem and related data.

    Returns
    -------
    master_soln : MasterResults
        Master solution results object.
    """
    config = master_data.config
    master_model = master_data.master_model
    master_soln = MasterResults(
        master_model=master_model, pyros_termination_condition=None
    )

    if config.solve_master_globally:
        solvers = [config.global_solver] + config.backup_global_solvers
    else:
        solvers = [config.local_solver] + config.backup_local_solvers

    solve_mode = "global" if config.solve_master_globally else "local"
    config.progress_logger.debug("Solving master problem")

    higher_order_decision_rule_efficiency(master_data)

    for idx, opt in enumerate(solvers):
        if idx > 0:
            config.progress_logger.debug(
                f"Invoking backup solver {opt!r} "
                f"(solver {idx + 1} of {len(solvers)}) for "
                f"master problem of iteration {master_data.iteration}."
            )
        results = call_solver(
            model=master_model,
            solver=opt,
            config=config,
            timing_obj=master_data.timing,
            timer_name="main.master",
            err_msg=(
                f"Optimizer {repr(opt)} ({idx + 1} of {len(solvers)}) "
                "encountered exception attempting to "
                f"solve master problem in iteration {master_data.iteration}"
            ),
        )

        master_soln.master_results_list.append(results)
        optimality_acceptable, infeasible = (
            process_termination_condition_master_problem(config=config, results=results)
        )
        time_out = check_time_limit_reached(master_data.timing, config)

        if optimality_acceptable:
            master_model.solutions.load_from(results)
            log_master_solve_results(master_model, config, results)
        if time_out:
            master_soln.pyros_termination_condition = pyrosTerminationCondition.time_out
        if infeasible:
            master_soln.pyros_termination_condition = (
                pyrosTerminationCondition.robust_infeasible
            )

        final_result_established = optimality_acceptable or time_out or infeasible
        if final_result_established:
            return master_soln

    # all solvers have failed to return an acceptable status.
    # we will terminate PyROS with subsolver error status.

    # log subproblem solve failure warning
    deterministic_model_qual = (
        " (i.e., the deterministic model)" if master_data.iteration == 0 else ""
    )
    deterministic_msg = (
        (
            " Please ensure that your deterministic model, "
            "subject to the nominal uncertain parameter realization "
            "you have provided, "
            f"is solvable by at least one of the subordinate {solve_mode} "
            "optimizers provided."
        )
        if master_data.iteration == 0
        else ""
    )
    master_soln.pyros_termination_condition = pyrosTerminationCondition.subsolver_error
    subsolver_termination_conditions = [
        res.solver.termination_condition for res in master_soln.master_results_list
    ]
    config.progress_logger.warning(
        f"Could not successfully solve master problem of iteration "
        f"{master_data.iteration}{deterministic_model_qual} with any of the "
        f"provided subordinate {solve_mode} optimizers. "
        f"(Termination statuses: "
        f"{[term_cond for term_cond in subsolver_termination_conditions]}.)"
        f"{deterministic_msg}"
    )

    # at this point, export subproblem to file, if desired.
    # NOTE: subproblem is written with variables set to their
    #       initial values (not the final subsolver iterate)
    if config.keepfiles and config.subproblem_file_directory is not None:
        write_subproblem(
            model=master_model,
            fname=(
                f"{config.uncertainty_set.type}"
                f"_{master_data.original_model_name}"
                f"_master_{master_data.iteration}"
            ),
            config=config,
        )

    return master_soln


def solve_master(master_data):
    """
    Solve the master problem.

    Returns
    -------
    master_soln : MasterResults
        Master problem solve results.
    """
    feasibility_problem_results = None
    time_out_after_feasibility = False
    if master_data.iteration > 0:
        feasibility_problem_results = solve_master_feasibility_problem(master_data)
        time_out_after_feasibility = check_time_limit_reached(
            master_data.timing, master_data.config
        )

    if time_out_after_feasibility:
        master_soln = MasterResults(
            master_model=master_data.master_model,
            feasibility_problem_results=feasibility_problem_results,
            master_results_list=None,
            pyros_termination_condition=pyrosTerminationCondition.time_out,
        )
    else:
        master_soln = solver_call_master(master_data)
        master_soln.feasibility_problem_results = feasibility_problem_results

    return master_soln


class MasterProblemData:
    """
    Container for objects pertaining to the PyROS master problem.

    Parameters
    ----------
    model_data : ModelData
        PyROS model data object, equipped with the
        fully preprocessed working model.

    Attributes
    ----------
    master_model : BlockData
        Master problem model object.
    original_model_name : str
        Name of the user-provided deterministic model object.
    iteration : int
        Index of the current PyROS cutting set iteration.
    timing : TimingData
        Main timer for the current problem being solved.
    config : ConfigDict
        PyROS solver options.
    """

    def __init__(self, model_data):
        """Initialize self (see docstring)."""
        self.master_model = construct_initial_master_problem(model_data)
        # we track the original model name for serialization purposes
        self.original_model_name = model_data.original_model.name
        self.iteration = 0
        self.timing = model_data.timing
        self.config = model_data.config

    def solve_master(self):
        """
        Solve the master problem.
        """
        return solve_master(self)

    def solve_dr_polishing(self):
        """
        Solve the DR polishing problem.
        """
        return minimize_dr_vars(self)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys

import pyomo
import pyomo.common.unittest as unittest

from pyomo.common.flags import NOTSET, in_testing_environment, building_documentation


class TestFlags(unittest.TestCase):

    def test_NOTSET(self):
        self.assertTrue(in_testing_environment())
        self.assertFalse(building_documentation())

        self.assertEqual(str(NOTSET), 'NOTSET')
        self.assertFalse(hasattr(pyomo, '__sphinx_build__'))
        self.assertEqual(repr(NOTSET), 'pyomo.common.flags.NOTSET')
        self.assertIsNone(in_testing_environment.state)

        try:
            pyomo.__sphinx_build__ = True
            self.assertTrue(in_testing_environment())
            self.assertTrue(building_documentation())
            self.assertEqual(repr(NOTSET), 'NOTSET')

            in_testing_environment(False)
            self.assertFalse(in_testing_environment())
            self.assertTrue(building_documentation())
            self.assertEqual(repr(NOTSET), 'NOTSET')
        finally:
            del pyomo.__sphinx_build__
            in_testing_environment(None)
        self.assertIsNone(in_testing_environment.state)

    def test_singleton(self):
        # This tests that the type is a "singleton", and that any
        # attempts to construct an instance will return the class
        self.assertIs(NOTSET(), NOTSET)
        self.assertIs(NOTSET(), NOTSET())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.set import (
    Reals,
    PositiveReals,
    NonPositiveReals,
    NegativeReals,
    NonNegativeReals,
    Integers,
    PositiveIntegers,
    NonPositiveIntegers,
    NegativeIntegers,
    NonNegativeIntegers,
    Boolean,
    Binary,
    Any,
    AnyWithNone,
    EmptySet,
    UnitInterval,
    PercentFraction,
    RealInterval,
    IntegerInterval,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import shutil
import glob
import os
import sys
import tempfile


def get_appsi_extension(in_setup=False, appsi_root=None):
    from pybind11.setup_helpers import Pybind11Extension

    if appsi_root is None:
        from pyomo.common.fileutils import this_file_dir

        appsi_root = this_file_dir()

    sources = [
        os.path.join(appsi_root, 'cmodel', 'src', file_)
        for file_ in (
            'interval.cpp',
            'expression.cpp',
            'common.cpp',
            'nl_writer.cpp',
            'lp_writer.cpp',
            'model_base.cpp',
            'fbbt_model.cpp',
            'cmodel_bindings.cpp',
        )
    ]

    if in_setup:
        package_name = 'pyomo.contrib.appsi.cmodel.appsi_cmodel'
    else:
        package_name = 'appsi_cmodel'
    if sys.platform.startswith('win'):
        # Assume that builds on Windows will use MSVC
        # MSVC doesn't have a flag for c++11, use c++14
        extra_args = ['/std:c++14']
    else:
        # Assume all other platforms are GCC-like
        extra_args = ['-std=c++11']
    return Pybind11Extension(package_name, sources, extra_compile_args=extra_args)


def build_appsi(args=[]):
    print('\n\n**** Building APPSI ****')
    from setuptools import Distribution
    from pybind11.setup_helpers import build_ext
    import pybind11.setup_helpers
    from pyomo.common.cmake_builder import handleReadonly
    from pyomo.common.envvar import PYOMO_CONFIG_DIR
    from pyomo.common.fileutils import this_file_dir

    class appsi_build_ext(build_ext):
        def run(self):
            basedir = os.path.abspath(os.path.curdir)
            if self.inplace:
                tmpdir = os.path.join(this_file_dir(), 'cmodel')
            else:
                tmpdir = os.path.abspath(tempfile.mkdtemp())
            print("Building in '%s'" % tmpdir)
            os.chdir(tmpdir)
            try:
                super(appsi_build_ext, self).run()
                if not self.inplace:
                    library = glob.glob("build/*/appsi_cmodel.*")[0]
                    target = os.path.join(
                        PYOMO_CONFIG_DIR,
                        'lib',
                        'python%s.%s' % sys.version_info[:2],
                        'site-packages',
                        '.',
                    )
                    if not os.path.exists(target):
                        os.makedirs(target)
                    shutil.copy(library, target)
            finally:
                os.chdir(basedir)
                if not self.inplace:
                    shutil.rmtree(tmpdir, onerror=handleReadonly)

    try:
        original_pybind11_setup_helpers_macos = pybind11.setup_helpers.MACOS
        pybind11.setup_helpers.MACOS = False

        package_config = {
            'name': 'appsi_cmodel',
            'packages': [],
            'ext_modules': [get_appsi_extension(False)],
            'cmdclass': {"build_ext": appsi_build_ext},
        }

        dist = Distribution(package_config)
        dist.script_args = ['build_ext'] + args
        dist.parse_command_line()
        dist.run_command('build_ext')
    finally:
        pybind11.setup_helpers.MACOS = original_pybind11_setup_helpers_macos


class AppsiBuilder:
    def __call__(self, parallel):
        return build_appsi()


if __name__ == '__main__':
    # Note: this recognizes the "--inplace" command line argument: build
    # directory will be put in the source tree (and preserved), and the
    # SO will be left in appsi/cmodel.
    build_appsi(sys.argv[1:])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
import logging

import pyomo.common.unittest as unittest

from pyomo.common.log import LoggingIntercept
from pyomo.common.tee import capture_output
from pyomo.core.expr import InequalityExpression
from pyomo.core.expr.sympy_tools import sympy_available
from pyomo.environ import (
    AbstractModel,
    BooleanConstant,
    BooleanVar,
    ConcreteModel,
    LogicalConstraint,
    LogicalConstraintList,
    TransformationFactory,
    Constraint,
    Param,
    NonNegativeIntegers,
)
from pyomo.repn import generate_standard_repn
from pyomo.gdp import Disjunction


class TestLogicalConstraint(unittest.TestCase):
    def create_model(self, abstract=False):
        if abstract is True:
            model = AbstractModel()
        else:
            model = ConcreteModel()
        model.x = BooleanVar()
        model.y = BooleanVar()
        model.z = BooleanVar()
        return model

    def test_abstract_and_empty(self):
        m = self.create_model(True)

        @m.LogicalConstraint()
        def p(m):
            return LogicalConstraint.Skip

        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot access property 'expr' on AbstractScalarLogicalConstraint "
            "'p' before it has been constructed",
        ):
            m.p.expr

        i = m.create_instance()
        with self.assertRaisesRegex(
            ValueError,
            "Accessing the expr of ScalarLogicalConstraint "
            "'p' before the LogicalConstraint has been assigned "
            "an expression.",
        ):
            i.p.body

    def test_construct(self):
        model = self.create_model()

        @model.LogicalConstraint()
        def p(model):
            return model.x

        self.assertIs(model.p.body, model.x)

        with LoggingIntercept() as LOG:
            with self.assertRaisesRegex(
                IndexError,
                "LogicalConstraint 'q': Cannot initialize multiple indices of "
                "a constraint with a single expression",
            ):
                model.q = LogicalConstraint(range(2), expr=model.x)
        self.assertEqual(
            LOG.getvalue().strip(),
            """
Rule failed when generating expression for LogicalConstraint q with index None:
IndexError: LogicalConstraint 'q': Cannot initialize multiple indices of a constraint with a single expression
Constructing component 'q' from data=None failed:
    IndexError: LogicalConstraint 'q': Cannot initialize multiple indices of a constraint with a single expression
""".strip(),
        )

        with LoggingIntercept(level=logging.DEBUG, module='pyomo.core.base') as LOG:
            model.r = LogicalConstraint()
        self.assertEqual(LOG.getvalue(), 'Constructing logical constraint r\n')

        with LoggingIntercept(level=logging.DEBUG, module='pyomo.core.base') as LOG:
            model.r.construct()
        self.assertEqual(LOG.getvalue(), "")

    def test_indexed_constructor(self):
        m = self.create_model()
        m.p = LogicalConstraint(
            range(7),
            rule=[
                LogicalConstraint.Feasible,
                LogicalConstraint.Infeasible,
                LogicalConstraint.Skip,
                m.x.implies(m.y),
                m.x,
                True,
                False,
            ],
        )

        self.assertExpressionsStructurallyEqual(m.p[0].expr, LogicalConstraint.Feasible)
        self.assertExpressionsStructurallyEqual(
            m.p[1].expr, LogicalConstraint.Infeasible
        )
        self.assertNotIn(2, m.p)
        with self.assertRaises(KeyError):
            m.p[2].expr
        self.assertExpressionsStructurallyEqual(m.p[3].expr, m.x.implies(m.y))
        self.assertExpressionsStructurallyEqual(m.p[4].expr, m.x)
        self.assertExpressionsStructurallyEqual(m.p[5].expr, BooleanConstant(True))
        self.assertExpressionsStructurallyEqual(m.p[6].expr, BooleanConstant(False))

    def test_display(self):
        m = self.create_model()
        m.p = LogicalConstraint(
            range(7),
            rule=[
                LogicalConstraint.Feasible,
                LogicalConstraint.Infeasible,
                LogicalConstraint.Skip,
                m.x.implies(m.y),
                m.x,
                True,
                False,
            ],
        )

        ref = """p : Size=6
    Key : Body
      0 :  True
      1 : False
      3 :  None
      4 :  None
      5 :  True
      6 : False
"""
        OUT = io.StringIO()
        m.p.display(ostream=OUT)
        self.assertEqual(ref, OUT.getvalue())

        with capture_output() as OUT:
            m.p.display()
        self.assertEqual(ref, OUT.getvalue())

        m.x = True
        m.y = False

        OUT = io.StringIO()
        m.p.display(ostream=OUT)
        self.assertEqual(
            """p : Size=6
    Key : Body
      0 :  True
      1 : False
      3 : False
      4 :  True
      5 :  True
      6 : False
""",
            OUT.getvalue(),
        )

        m.p.deactivate()

        OUT = io.StringIO()
        m.p.display(ostream=OUT)
        self.assertEqual("", OUT.getvalue())

    def test_pprint(self):
        m = self.create_model()
        m.p = LogicalConstraint(
            range(7),
            rule=[
                LogicalConstraint.Feasible,
                LogicalConstraint.Infeasible,
                LogicalConstraint.Skip,
                m.x.implies(m.y),
                m.x,
                True,
                False,
            ],
        )

        OUT = io.StringIO()
        m.p.pprint(ostream=OUT)
        self.assertEqual(
            """p : Size=6, Index={0, 1, 2, 3, 4, 5, 6}, Active=True
    Key : Body       : Active
      0 :   Feasible :   True
      1 : Infeasible :   True
      3 :    x --> y :   True
      4 :          x :   True
      5 :       True :   True
      6 :      False :   True
""",
            OUT.getvalue(),
        )

    def test_indexed_deferred_constructor(self):
        m = self.create_model()

        def p_rule(m, i):
            return [
                LogicalConstraint.Feasible,
                LogicalConstraint.Infeasible,
                LogicalConstraint.Skip,
                m.x.implies(m.y),
                m.x,
                True,
                False,
            ][i]

        m.p = LogicalConstraint(NonNegativeIntegers, rule=p_rule)

        self.assertExpressionsStructurallyEqual(m.p[0].expr, LogicalConstraint.Feasible)
        self.assertExpressionsStructurallyEqual(
            m.p[1].expr, LogicalConstraint.Infeasible
        )
        self.assertNotIn(2, m.p)
        with self.assertRaises(KeyError):
            m.p[2].expr
        self.assertExpressionsStructurallyEqual(m.p[3].expr, m.x.implies(m.y))
        self.assertExpressionsStructurallyEqual(m.p[4].expr, m.x)
        self.assertExpressionsStructurallyEqual(m.p[5].expr, BooleanConstant(True))
        self.assertExpressionsStructurallyEqual(m.p[6].expr, BooleanConstant(False))

        m.q = LogicalConstraint(NonNegativeIntegers)
        with self.assertRaises(KeyError):
            m.q[2].expr

    def test_add(self):
        m = self.create_model()
        m.p = LogicalConstraint()
        expr = m.x.implies(m.y)
        with self.assertRaisesRegex(
            ValueError,
            "ScalarLogicalConstraint object 'p' does not accept "
            "index values other than None. Invalid value: 1",
        ):
            m.p.add(1, expr)
        self.assertEqual(len(m.p), 0)
        m.p.add(None, expr)
        self.assertEqual(len(m.p), 1)
        self.assertIs(m.p.expr, expr)

        m.q = LogicalConstraint(range(2))
        m.q.add(0, expr)
        self.assertEqual(len(m.q), 1)
        self.assertIs(m.q[0].expr, expr)
        m.q.add(1, expr)
        self.assertEqual(len(m.q), 2)
        self.assertIs(m.q[1].expr, expr)

    def test_set_value(self):
        m = self.create_model()

        m.p = LogicalConstraint()
        self.assertEqual(len(m.p), 0)

        expr = m.x.implies(m.y)
        m.p = expr
        self.assertIs(m.p.expr, expr)

        with self.assertRaisesRegex(
            ValueError, "LogicalConstraint 'p': rule returned None"
        ):
            m.p = None
        # (test that the expr is unchanged on error)
        self.assertIs(m.p.expr, expr)

        m.p = LogicalConstraint.Skip
        self.assertEqual(len(m.p), 0)

        m.p = LogicalConstraint.Feasible
        self.assertExpressionsStructurallyEqual(m.p.expr, LogicalConstraint.Feasible)

        m.p = LogicalConstraint.Infeasible
        self.assertExpressionsStructurallyEqual(m.p.expr, LogicalConstraint.Infeasible)

        with self.assertRaisesRegex(
            ValueError, "Assigning improper value to LogicalConstraint 'p'."
        ):
            m.p = LogicalConstraint

        with self.assertRaisesRegex(
            ValueError, "Assigning improper value to LogicalConstraint 'p'."
        ):
            m.p = {}

        with self.assertRaisesRegex(
            ValueError, "Assigning improper value to LogicalConstraint 'p'."
        ):
            m.p = Param(mutable=True) + 1

    def test_get_value(self):
        m = self.create_model()
        m.p = LogicalConstraint(expr=m.x.implies(m.y))
        self.assertIs(m.p.get_value(), m.p.expr)

    def check_lor_on_disjunct(self, model, disjunct, x1, x2):
        x1 = x1.get_associated_binary()
        x2 = x2.get_associated_binary()
        disj0 = disjunct.logic_to_linear
        self.assertEqual(len(disj0.component_map(Constraint)), 1)
        lor = disj0.transformed_constraints[1]
        self.assertEqual(lor.lower, 1)
        self.assertIsNone(lor.upper)
        repn = generate_standard_repn(lor.body)
        self.assertEqual(repn.constant, 0)
        self.assertTrue(repn.is_linear())
        self.assertEqual(len(repn.linear_vars), 2)
        self.assertIs(repn.linear_vars[0], x1)
        self.assertIs(repn.linear_vars[1], x2)
        self.assertEqual(repn.linear_coefs[0], 1)
        self.assertEqual(repn.linear_coefs[1], 1)

    @unittest.skipUnless(sympy_available, "Sympy not available")
    def test_statement_in_Disjunct_with_logical_to_linear(self):
        # This is an old test that originally tested that GDP's
        # BigM/Hull correctly handled Disjuncts with LogicalConstraints
        # (implicitly calling logical_to_linear to leave the transformed
        # algebraic constraints on the Disjuncts).  That is no longer
        # the default behavior.  However, we will preserve this (with an
        # explicit call to logical_to_linear) for posterity
        model = self.create_model()
        model.disj = Disjunction(expr=[[model.x.lor(model.y)], [model.y.lor(model.z)]])

        TransformationFactory('core.logical_to_linear').apply_to(
            model, targets=model.disj.disjuncts
        )

        bigmed = TransformationFactory('gdp.bigm').create_using(model)
        # check that the algebraic versions are living on the Disjuncts
        self.check_lor_on_disjunct(bigmed, bigmed.disj.disjuncts[0], bigmed.x, bigmed.y)
        self.check_lor_on_disjunct(bigmed, bigmed.disj.disjuncts[1], bigmed.y, bigmed.z)

        TransformationFactory('gdp.hull').apply_to(model)
        self.check_lor_on_disjunct(model, model.disj.disjuncts[0], model.x, model.y)
        self.check_lor_on_disjunct(model, model.disj.disjuncts[1], model.y, model.z)

    # TODO look to test_con.py for inspiration
    def test_deprecated_rule_attribute(self):
        def rule(m):
            return m.x.implies(m.x)

        def new_rule(m):
            return m.x.implies(~m.x)

        m = ConcreteModel()
        m.x = BooleanVar()
        m.con = LogicalConstraint(rule=rule)

        self.assertIs(m.con.rule._fcn, rule)
        with LoggingIntercept() as LOG:
            m.con.rule = new_rule
        self.assertIn(
            "DEPRECATED: The 'LogicalConstraint.rule' attribute will be made "
            "read-\nonly",
            LOG.getvalue(),
        )
        self.assertIs(m.con.rule, new_rule)


class TestLogicalConstraintList(unittest.TestCase):
    def create_model(self, abstract=False):
        if abstract is True:
            model = AbstractModel()
        else:
            model = ConcreteModel()
        model.x = BooleanVar()
        model.y = BooleanVar()
        model.z = BooleanVar()
        return model

    def test_construct(self):
        m = self.create_model()

        @m.LogicalConstraintList(starting_index=0)
        def p(m):
            yield m.x
            yield m.x.implies(m.y)
            yield True
            yield LogicalConstraintList.Skip
            yield False
            yield LogicalConstraintList.End
            yield m.y

        self.assertEqual(len(m.p), 4)
        self.assertExpressionsStructurallyEqual(m.p[0].expr, m.x)
        self.assertExpressionsStructurallyEqual(m.p[1].expr, m.x.implies(m.y))
        self.assertExpressionsStructurallyEqual(m.p[2].expr, BooleanConstant(True))
        self.assertExpressionsStructurallyEqual(m.p[3].expr, BooleanConstant(False))

        with self.assertRaisesRegex(
            ValueError, "LogicalConstraintList does not accept the 'expr' keyword"
        ):
            m.q = LogicalConstraintList(expr=[m.x])

        with LoggingIntercept(level=logging.DEBUG, module='pyomo.core.base') as LOG:
            m.r = LogicalConstraintList()
        self.assertEqual(LOG.getvalue(), 'Constructing logical constraint list r\n')

        with LoggingIntercept(level=logging.DEBUG, module='pyomo.core.base') as LOG:
            m.r.construct()
        self.assertEqual(LOG.getvalue(), "")


if __name__ == "__main__":
    unittest.main()
